# scurrypy/core

from .error import DiscordError
from .events import EVENTS
from .gateway import GatewayClient
from .http import HTTPClient
from .model import DataModel
from .snowflake import Snowflake

__all__ = [
    "DiscordError",
    "EVENTS",
    "GatewayClient",
    "HTTPClient",
    "DataModel",
    "Snowflake"
]



from typing import TypedDict, Optional

from ..enums.guild import OnboardingMode

from ..api.guilds.welcome_screen import WelcomeScreenChannelPart
from ..api.guilds.onboarding import OnboardingPromptPart
from ..api.guilds.role import GuildRoleColorsPart
from ..api.image_data import ImageDataPart

class EditGuildParams(TypedDict, total=False):
    """Represents fields for editing a guild."""

    name: str
    """Guild name."""

    afk_channel_id: int
    """Channel ID for AFK channel."""

    icon: ImageDataPart
    """Data URI scheme for guild icon.
    
    !!! note
        Can be animated if guild has `ANIMATED_ICON` feature.
    """

    splash: ImageDataPart
    """Data URI scheme for guild splash if guild has `INVITE_SPLASH` feature."""

    discovery_splash: ImageDataPart
    """Data URI scheme for guild discovery if guild has `DISCOVERABLE` feature."""

    banner: ImageDataPart
    """Data URI scheme for guild banner if guild has `BANNER` feature.
    
    !!! note
        Can be animated if guild has `ANIMATED_BANNER` feature.
    """

    system_channel_id: int
    """Channel ID for receiving guild notices (e.g., boosts, user join)."""

    rules_channel_id: int
    """Channel ID for where guilds display rules."""

    public_updates_channel_id: int
    """Channel ID for receiving notices from Discord."""

    description: str
    """Description for the guild."""

    premium_progress_bar_enabled: bool
    """Whether the guild's boost progress bar should be shown."""

    safety_alerts_channel_id: int
    """Channel ID for receiving safety alerts from Discord."""

class EditGuildRoleParams(TypedDict, total=False):
    """Represents fields for editing a guild role."""

    name: Optional[str]
    """Name of the role."""

    colors: Optional[GuildRoleColorsPart]
    """Colors of the role."""

    hoist: Optional[bool]
    """Whether the role is displayed separately on the sidebar."""

    icon: Optional[ImageDataPart]
    """Icon of the role (if guild has `ROLE_ICONS` feature)."""

    permissions: int
    """Permission bit set. [`INT_LIMIT`]"""

    unicode_emoji: Optional[str]
    """Unicode emoji of the role (if guilde has `ROLE_ICONS` feature)."""

    mentionable: Optional[bool]
    """Whether the role should be mentionable."""

class EditGuildWelcomeScreenParams(TypedDict, total=False):
    """Represents fields for editing a guild welcome screen."""

    enabled: bool
    """Whether the welcome scren is enabled."""

    welcome_channels: list[WelcomeScreenChannelPart]
    """Channels linked when the welcome screen is displayed."""

    description: str
    """Guild description to show on the welcome screen."""

class EditOnboardingParams(TypedDict, total=False):
    """Represents fields for editing a guild onboarding flow."""

    prompts: list[OnboardingPromptPart]
    """Prompts shown during onboarding."""

    default_channel_ids: list[int]
    """Channel IDs that members get opted into automatically."""

    enabled: bool
    """Whether onboarding is enabled in the guild."""
    
    mode: OnboardingMode
    """Current mode of onboarding."""

class EditGuildStickerParams(TypedDict, total=False):
    """Represents fields for editing a guild sticker."""

    name: str
    """Name of the sticker."""

    description: str
    """Description of the sticker."""

    tags: str
    """Autocomplete/suggestion tags for the sticker."""



from .enum_types import DiscordFlags

class ApplicationFlags(DiscordFlags):
    """Application flags."""

    GATEWAY_PRESENCE = 1 << 12
    """Privileged intent to receive presence_update events."""

    GATEWAY_PRESENCE_LIMITED = 1 << 13
    """Intent to receive presence_update events."""

    GATEWAY_GUILD_MEMBERS = 1 << 14
    """Privileged intent to receive member-related events."""

    GATEWAY_GUILD_MEMBERS_LIMITED = 1 << 15
    """Intent to receive member-related events."""

    VERIFICATION_PENDING_GUILD_LIMIT = 1 << 16
    """Indicates unusual growth of an app that prevents verification."""

    GATEWAY_MESSAGE_CONTENT = 1 << 18
    """Privileged intent to receive message content."""

    GATEWAY_MESSAGE_CONTENT_LIMITED = 1 << 19
    """Intent to receive message content."""



from dataclasses import dataclass
from typing import Unpack

from .base_resource import BaseResource

from ..core.snowflake import Snowflake

from ..api.emoji import EmojiModel, ApplicationEmojiPart, GuildEmojiPart

from ..params.emoji import EditApplicationEmojiParams, EditGuildEmojiParams

@dataclass
class ApplicationEmoji(BaseResource):
    """Represents a Discord Bot Emoji."""

    application_id: Snowflake
    """Application ID of the emojis."""

    async def fetch(self, emoji_id: Snowflake) -> EmojiModel:
        """Fetch an emoji from the bot repository.

        Args:
            emoji_id (Snowflake): emoji ID

        Returns:
            (EmojiModel): queried emoji
        """
        data = await self.http.request("GET", f"/applications/{self.application_id}/emojis/{emoji_id}")

        return EmojiModel.from_dict(data)
    
    async def fetch_all(self) -> list[EmojiModel]:
        """Fetch all emojis from the bot repository.

        Returns:
            (list[EmojiModel]): queried list of bot emojis
        """
        data = await self.http.request("GET", f"/applications/{self.application_id}/emojis")

        emojis = data.get("items")

        return [EmojiModel.from_dict(emoji) for emoji in emojis]
    
    async def create(self, emoji: ApplicationEmojiPart) -> EmojiModel:
        """Add an emoji to the bot emoji repository.

        Args:
            emoji (ApplicationEmojiPart): bot emoji fields

        Returns:
            (EmojiModel): new emoji
        """
        data = await self.http.request(
            'POST', 
            f'/applications/{self.application_id}/emojis',
            data=emoji.to_dict()
        )
    
        return EmojiModel.from_dict(data)
    
    async def edit(self, emoji_id: Snowflake, **options: Unpack[EditApplicationEmojiParams]) -> EmojiModel:
        """Edit an emoji in the bot repository.

        Args:
            emoji_id (Snowflake): ID of the emoji
            options (EditBotEmojiParams): fields to edit the emoji

        Returns:
            (EmojiModel): updated emoji
        """
        data = await self.http.request(
            'PATCH', 
            f'/applications/{self.application_id}/emojis/{emoji_id}', 
            data=options
        )

        return EmojiModel.from_dict(data)

    async def delete(self, emoji_id: Snowflake) -> None:
        """Deletes an emoji from the bot repository.

        Args:
            emoji_id (int): ID of the emoji to remove
        """
        await self.http.request('DELETE', f'/applications/{self.application_id}/emojis/{emoji_id}')

@dataclass
class GuildEmoji(BaseResource):
    """Represents a Discord Guild Emoji."""

    guild_id: Snowflake
    """Guild ID of the emojis."""

    async def fetch(self, emoji_id: Snowflake) -> EmojiModel:
        """Fetch an emoji from this guild.

        Args:
            emoji_id (Snowflake): emoji ID

        Returns:
            (EmojiModel): queried guild emoji
        """
        data = await self.http.request("GET", f"/guilds/{self.guild_id}/emojis/{emoji_id}")

        return EmojiModel.from_dict(data)
    
    async def fetch_all(self) -> list[EmojiModel]:
        """Fetch all emojis from this guild.

        Returns:
            (list[EmojiModel]): queried list of guild emojis
        """
        data = await self.http.request("GET", f"/guilds/{self.guild_id}/emojis")

        return [EmojiModel.from_dict(emoji) for emoji in data]

    async def create(self, emoji: GuildEmojiPart) -> EmojiModel:
        """Create a new emoji for this guild.
        Fires [`GuildEmojisUpdateEvent`][scurrypy.events.guild_events.GuildEmojisUpdateEvent].

        Args:
            emoji (GuildEmojiPart): fields for creating a guild emoji

        Returns:
            (EmojiModel): new emoji
        """
        data = await self.http.request(
            'POST', 
            f'/guilds/{self.guild_id}/emojis', 
            data=emoji.to_dict()
        )

        return EmojiModel.from_dict(data)
    
    async def edit(self, emoji_id: Snowflake, **options: Unpack[EditGuildEmojiParams]) -> EmojiModel:
        """Edit a guild emoji in this guild.
        Fires [`GuildEmojisUpdateEvent`][scurrypy.events.guild_events.GuildEmojisUpdateEvent].

        Args:
            emoji_id (Snowflake): ID of the emoji to edit
            options (EditGuildEmojiParams): params for editing a guild's emoji

        Returns:
            (EmojiModel): updated emoji
        """
        data = await self.http.request(
            'PATCH', 
            f'/guilds/{self.guild_id}/emojis/{emoji_id}', 
            data=options
        )

        return EmojiModel.from_dict(data)

    async def delete(self, emoji_id: Snowflake) -> None:
        """Delete an emoji from this guild.
        Fires [`GuildEmojisUpdateEvent`][scurrypy.events.guild_events.GuildEmojisUpdateEvent].

        !!! important "Permissions"
            * `CREATE_GUILD_EXPRESSIONS` → required if created by the current user (or `MANAGE_GUILD_EXPRESSIONS`)
            * `MANAGE_GUILD_EXPRESSIONS` → required for other emojis

        Args:
            emoji_id (Snowflake): ID of the emoji
        """
        await self.http.request('DELETE', f'/guilds/{self.guild_id}/emojis/{emoji_id}')



from dataclasses import dataclass
from typing import Unpack

from .base_resource import BaseResource

from ..core.snowflake import Snowflake
from ..core.serialization import serialize

from ..enums.message import MessageFlags
from ..enums.emoji import ReactionType

from ..api.messages.message import MessageModel

from ..api.emoji import EmojiModel
from ..api.user import UserModel

from ..params.message import EditMessageParams

class _EditMessageMixin:
    """Common message edit methods."""

    def _apply_suppress_embeds(
        self,
        payload: dict,
        suppress_embeds: bool
    ) -> None:
        if suppress_embeds is not None:
            flags = payload.get("flags", 0)

            if suppress_embeds:
                flags |= MessageFlags.SUPPRESS_EMBEDS
            else:
                flags &= ~MessageFlags.SUPPRESS_EMBEDS

            payload["flags"] = flags

    def _prepare_attachments(
        self,
        payload: dict
    ) -> list[str] | None:
        if "attachments" not in payload:
            return None

        attachments = payload["attachments"]

        for idx, attachment in enumerate(attachments):
            attachment.id = idx

        payload["attachments"] = [
            attachment.to_dict()
            for attachment in attachments
        ]

        return [attachment.path for attachment in attachments]

@dataclass
class Message(BaseResource, _EditMessageMixin):
    """A Discord message."""

    id: Snowflake
    """ID of the message"""

    channel_id: Snowflake
    """Channel ID of the message."""

    async def fetch(self) -> MessageModel:
        """Fetches the message data based on the given channel ID and message id.

        Returns:
            (MessageModel): queried message
        """
        data = await self.http.request('GET', f"/channels/{self.channel_id}/messages/{self.id}")

        return MessageModel.from_dict(data)
    
    async def edit(
        self,
        *,
        suppress_embeds: bool = None,
        **options: Unpack[EditMessageParams]
    ) -> MessageModel:
        """Edits this message.
        Fires [`MessageUpdateEvent`][scurrypy.events.message_events.MessageUpdateEvent].

        !!! important "Permissions"
            Requires `MANAGE_MESSAGES` *only* if editing another user's message or to edit flags

        Args:
            options (EditMessageParams): fields to edit for the message
            suppress_embeds (optional, bool): whether the response's embeds should be removed

        Returns:
            (MessageModel): updated message
        """
        
        options = serialize(options)
        self._apply_suppress_embeds(options, suppress_embeds)
        files = self._prepare_attachments(options)

        data = await self.http.request(
            "PATCH",
            f"/channels/{self.channel_id}/messages/{self.id}",
            data=options,
            files=files,
        )

        return MessageModel.from_dict(data)

    async def crosspost(self) -> MessageModel:
        """Crosspost this message in an Annoucement channel to all following channels.
        Fires [`MessageUpdateEvent`][scurrypy.events.message_events.MessageUpdateEvent].

        !!! important "Permissions"
            * `SEND_MESSAGES` → required to publish your own messages
            * `MANAGE_MESSAGES` → required to publish messages from others

        Returns:
            (MessageModel): published (crossposted) message
        """
        data = await self.http.request('POST', f'/channels/{self.channel_id}/messages/{self.id}/crosspost')

        return MessageModel.from_dict(data)

    async def delete(self):
        """Deletes this message.
        Fires [`MessageDeleteEvent`][scurrypy.events.message_events.MessageDeleteEvent].
        """
        await self.http.request("DELETE", f"/channels/{self.channel_id}/messages/{self.id}")

    async def pin(self) -> None:
        """Pin this message to its channel's pins.
        Fires [`ChannelPinsUpdateEvent`][scurrypy.events.channel_events.ChannelPinsUpdateEvent].
        """
        await self.http.request('PUT', f'/channels/{self.channel_id}/messages/pins/{self.id}')
    
    async def unpin(self) -> None:
        """Unpin this message from its channel's pins.
        Fires [`ChannelPinsUpdateEvent`][scurrypy.events.channel_events.ChannelPinsUpdateEvent].
        """
        await self.http.request('DELETE', f'/channels/{self.channel_id}/messages/pins/{self.id}')

    async def fetch_emoji_reactions(self, 
        emoji: EmojiModel | str, 
        type: ReactionType = ReactionType.NORMAL, 
        after: int = None, 
        limit: int = 25
    ) -> list[UserModel]:
        """Fetches users who reacted with the specified emoji parameters.

        Args:
            emoji (EmojiModel | str): the standard emoji (str) or custom emoji (EmojiModel)
            type (ReactionType, optional): Type of emoji. Defaults to `ReactionType.NORMAL`.
            after (int, optional): users after this ID
            limit (int, optional): Max number of users to return. Defaults to `25`.

        Returns:
            list[UserModel]: list of users who reacted with this emoji
        """
        if isinstance(emoji, str):
            emoji = EmojiModel(emoji)

        data = await self.http.request(
            'GET',
            f"/channels/{self.channel_id}/messages/{self.id}/reactions/{emoji.api_code}",
            params={
                'type': type,
                'after': after,
                'limit': limit
            }
        )
        return [UserModel.from_dict(user) for user in data]

    async def add_reaction(self, emoji: EmojiModel | str) -> None:
        """Add a reaction to this message.
        Fires [`MessageReactionAddEvent`][scurrypy.events.reaction_events.ReactionAddEvent].

        !!! important "Permissions"
            Requires `READ_MESSAGE_HISTORY` and `ADD_REACTIONS`

        Args:
            emoji (EmojiModel | str): the standard emoji (str) or custom emoji (EmojiModel)
        """
        if isinstance(emoji, str):
            emoji = EmojiModel(emoji)

        await self.http.request(
            "PUT",
            f"/channels/{self.channel_id}/messages/{self.id}/reactions/{emoji.api_code}/@me")

    async def remove_reaction(self, emoji: EmojiModel | str) -> None:
        """Remove the bot's reaction from this message.
        Fires [`MessageReactionRemoveEvent`][scurrypy.events.reaction_events.ReactionRemoveEvent].

        Args:
            emoji (EmojiModel | str): the standard emoji (str) or custom emoji (EmojiModel)
        """
        if isinstance(emoji, str):
            emoji = EmojiModel(emoji)

        await self.http.request(
            "DELETE",
            f"/channels/{self.channel_id}/messages/{self.id}/reactions/{emoji.api_code}/@me")

    async def remove_user_reaction(self, emoji: EmojiModel | str, user_id: Snowflake) -> None:
        """Remove a specific user's reaction from this message.
        Fires [`MessageReactionRemoveEvent`][scurrypy.events.reaction_events.ReactionRemoveEvent].

        !!! important "Permissions"
            Requires `MANAGE_MESSAGES`

        Args:
            emoji (EmojiModel | str): the standard emoji (str) or custom emoji (EmojiModel)
            user_id (Snowflake): user's ID
        """
        if isinstance(emoji, str):
            emoji = EmojiModel(emoji)

        await self.http.request(
            "DELETE",
            f"/channels/{self.channel_id}/messages/{self.id}/reactions/{emoji.api_code}/{user_id}")

    async def remove_emoji_reaction(self, emoji: EmojiModel | str) -> None:
        """Clear all reactions for a given emoji from this message.

        !!! important "Permissions"
            Requires `MANAGE_MESSAGES`

        Args:
            emoji (EmojiModel | str): the standard emoji (str) or custom emoji (EmojiModel)
        """
        if isinstance(emoji, str):
            emoji = EmojiModel(emoji)
        
        await self.http.request(
            "DELETE",
            f"/channels/{self.channel_id}/messages/{self.id}/reactions/{emoji.api_code}")

    async def remove_all_reactions(self) -> None:
        """Clear all reactions from this message.
        Fires [`MessageReactionRemoveAllEvent`][scurrypy.events.reaction_events.ReactionRemoveAllEvent].

        !!! important "Permissions"
            Requires `MANAGE_MESSAGES`
        """
        await self.http.request(
            "DELETE",
            f"/channels/{self.channel_id}/messages/{self.id}/reactions")



# scurrypy/api/guilds

from .ban import (
    GuildBanModel, 
    BulkGuildBanModel, 
    BulkGuildBanPart
)
from .guild import (
    ReadyGuildModel, 
    UnavailableGuildModel, 
    GuildModel
)
from .onboarding import (
    OnboardingPromptOptionModel, 
    OnboardingPromptModel, 
    GuildOnboadingModel, 
    OnboardingPromptOptionPart, 
    OnboardingPromptPart
)
from .role import (
    GuildRoleColorModel, 
    GuildRoleModel, 
    GuildRoleColorsPart, 
    GuildRolePart
)
from .welcome_screen import (
    GuildWelcomeChannelModel, 
    GuildWelcomeScreenModel, 
    WelcomeScreenChannelPart
)

__all__ = [
    "GuildBanModel", 
    "BulkGuildBanModel", 
    "BulkGuildBanPart",

    "ReadyGuildModel", 
    "UnavailableGuildModel", 
    "GuildModel",

    "OnboardingPromptOptionModel", 
    "OnboardingPromptModel", 
    "GuildOnboadingModel", 
    "OnboardingPromptOptionPart", 
    "OnboardingPromptPart",

    "GuildRoleColorModel", 
    "GuildRoleModel", 
    "GuildRoleColorsPart", 
    "GuildRolePart",

    "GuildWelcomeChannelModel", 
    "GuildWelcomeScreenModel", 
    "WelcomeScreenChannelPart"
]



from ...enums.components import ComponentType

from .button import Button
from .layout import ActionRow, Section, TextDisplay, Thumbnail, MediaGallery, File, Separator, Container, Label
from .modal import TextInput, FileUpload, RadioGroup, CheckboxGroup, Checkbox
from .select_menu import StringSelect, UserSelect, RoleSelect, MentionableSelect, ChannelSelect

COMPONENT_MAP = {
    ComponentType.ACTION_ROW: ActionRow,
    ComponentType.BUTTON: Button,
    ComponentType.STRING_SELECT: StringSelect,
    ComponentType.TEXT_INPUT: TextInput,
    ComponentType.USER_SELECT: UserSelect,
    ComponentType.ROLE_SELECT: RoleSelect,
    ComponentType.MENTIONABLE_SELECT: MentionableSelect,
    ComponentType.CHANNEL_SELECT: ChannelSelect,
    ComponentType.SECTION: Section,
    ComponentType.TEXT_DISPLAY: TextDisplay,
    ComponentType.THUMBNAIL: Thumbnail,
    ComponentType.MEDIA_GALLERY: MediaGallery,
    ComponentType.FILE: File,
    ComponentType.SEPARATOR: Separator,
    ComponentType.CONTAINER: Container,
    ComponentType.LABEL: Label,
    ComponentType.FILE_UPLOAD: FileUpload,
    ComponentType.RADIO_GROUP: RadioGroup,
    ComponentType.CHECKBOX_GROUP: CheckboxGroup,
    ComponentType.CHECKBOX: Checkbox
}

class MessageComponentFactory:
    """Represents top-level components."""

    @staticmethod
    def from_dict(data: dict):
        component_type = ComponentType(int(data["type"]))
        model_cls = COMPONENT_MAP.get(component_type)

        if not model_cls:
            raise ValueError(f"Invalid component type: {component_type}")

        return model_cls.from_dict(data)



# scurrypy/api/components

from .button import Button
from .layout import (
    ActionRow,
    Section, 
    TextDisplay, 
    Thumbnail, 
    MediaGalleryItem, 
    MediaGallery, 
    File, 
    Separator, 
    Container,
    Label
)
from .modal import (
    TextInput, 
    FileUpload, 
    ListOption, 
    RadioGroup, 
    CheckboxGroup, 
    Checkbox
)
from .select_menu import (
    SelectOption, 
    StringSelect, 
    DefaultValue, 
    SelectMenu, 
    UserSelect, 
    RoleSelect, 
    MentionableSelect, 
    ChannelSelect
)
from .factory import MessageComponentFactory

__all__ = [
    "MessageComponentFactory",
    
    "Button",

    "ActionRow",
    "Section", 
    "TextDisplay", 
    "Thumbnail", 
    "MediaGalleryItem", 
    "MediaGallery", 
    "File", 
    "Separator", 
    "Container",
    "Label",

    "TextInput", 
    "FileUpload", 
    "ListOption", 
    "RadioGroup", 
    "CheckboxGroup", 
    "Checkbox",

    "SelectOption", 
    "StringSelect", 
    "DefaultValue", 
    "SelectMenu", 
    "UserSelect", 
    "RoleSelect", 
    "MentionableSelect", 
    "ChannelSelect"
]



# scurrypy/bases

from .addon import Addon
from .channel import GuildChannelCreate
from .components import (
    Component,
    ActionRowChild,
    SectionChild,
    SectionAccessoryChild,
    ContainerChild,
    LabelChild
)
from .interaction import InteractionData

__all__ = [
    "Addon",

    "GuildChannelCreate",

    "Component",
    "ActionRowChild",
    "SectionChild",
    "SectionAccessoryChild",
    "ContainerChild",
    "LabelChild",

    "InteractionData"
]



from dataclasses import dataclass, field

from .base_event import Event

from ..bases.interaction import InteractionData

from ..enums.interaction import InteractionType

from ..api.interactions.interaction import InteractionModel, ApplicationCommandDataModel, MessageComponentDataModel, ModalDataModel

@dataclass
class InteractionEvent(Event, InteractionModel):

    data: InteractionData = field(init=False, default=None)
    """Interaction response data. Can be one of `InteractionData`[scurrypy.bases.InteractionData]'s variants."""

    @classmethod
    def from_dict(cls, data: dict):

        obj = super().from_dict(data) # InteractionModel's DataModel

        interaction_data = data.get("data")
        interaction_type = data.get("type")

        match interaction_type:
            case InteractionType.APPLICATION_COMMAND | InteractionType.APPLICATION_COMMAND_AUTOCOMPLETE:
                obj.data = ApplicationCommandDataModel.from_dict(interaction_data)
            case InteractionType.MESSAGE_COMPONENT:
                obj.data = MessageComponentDataModel.from_dict(interaction_data)
            case InteractionType.MODAL_SUBMIT:
                obj.data = ModalDataModel.from_dict(interaction_data)
        
        return obj



from typing import TypedDict, Optional, Literal

from ..enums.channel import ChannelType, ChannelFlags, SortOrderType, ForumLayoutType

from ..api.channels.default_reaction import DefaultReactionPart
from ..api.channels.tag import TagPart

class EditGuildChannelParams(TypedDict, total=False):
    """Parameters for editing a guild channel."""

    name: str
    """Name of the channel."""

    type: Optional[ChannelType]
    """Type of channel.
    
    !!! important
        Only conversion between text and announcement is supported in guilds with `NEWS` feature.
    """

    position: Optional[int]
    """Sorting position of the channel (channels with the same position are sorted by id)."""

    topic: Optional[str]
    """Topic of the channel."""

    nsfw: Optional[bool]
    """If the channel is flagged NSFW."""

    rate_limit_per_user: Optional[int]
    """Seconds user must wait between sending messages in the channel."""

    parent_id: Optional[int]
    """Category ID of the channel."""

    default_auto_archive_duration: Optional[int]
    """Default duration in minutes threads will be hidden after period of inactivity."""

    flags: Optional[ChannelFlags]
    """Channel flags."""

    default_reaction_emoji: Optional[DefaultReactionPart]
    """Emoji to show in the add reaction button in a `GUILD_FORUM` post."""

    available_tags: Optional[list[TagPart]]
    """Set of tags that can be applied to a `GUILD_FORUM` post."""

    default_sort_order: Optional[SortOrderType]
    """Default forum sort order."""

    default_forum_layout: Optional[ForumLayoutType]
    """Default forum layout view."""

    default_thread_rate_limit_per_user: Optional[int]
    """Rate limit per user set on newly created threads.
    
    !!! note
        This field does not live update!
    """

class EditThreadChannelParams(TypedDict, total=False):
    """Parameters for editing a thread channel."""

    name: Optional[str]
    """Name of the channel."""

    archived: Optional[bool]
    """Whether the thread is archived."""

    auto_archive_duration: Optional[Literal[60, 1440, 4320, 10080]]
    """Duration in minutes threads will be hidden after period of inactivity."""

    locked: bool
    """Whether the thread is locked.
    
    !!! note
        Only users with `MANAGE_THREADS` can unarchive the thread.
    """

    invitable: Optional[bool]
    """Whether non-moderators can add other non-moderators to the thread (private threads only)."""

    rate_limit_per_user: Optional[int]
    """Seconds user must wait between sending messages in the channel."""

    flags: Optional[ChannelFlags]
    """Channel flags."""

    applied_tags: Optional[list[int]]
    """Set of tags applied to a `GUILD_FORUM` post."""



from .enum_types import DiscordFlags, DiscordTypes

class MessageReferenceType(DiscordTypes):
    """Constants associated with how reference data is populated."""

    DEFAULT = 0
    """Standard reference used by replies."""

    FORWARD = 1
    """Reference used to point to a message at a point in time."""

class MessageFlags(DiscordFlags):
    """Flags that can be applied to a message."""

    NO_FLAGS = 0
    """Message has no flags."""

    CROSSPOSTED = 1 << 0
    """Message has been published."""

    IS_CROSSPOST = 1 << 1
    """Message originated from another channel."""

    SUPPRESS_EMBEDS = 1 << 2
    """Hide embeds."""

    EPHEMERAL = 1 << 6
    """Only visible to the invoking user."""

    IS_COMPONENTS_V2 = 1 << 15
    """This message includes Discord's V2 Components."""

class MessageType(DiscordTypes):
    """Tyoes if messages."""

    DEFAULT = 0
    CHANNEL_PINNED_MESSAGE = 4
    GUILD_BOOST = 8
    GUILD_BOOST_TIER_1 = 9
    GUILD_BOOST_TIER_2 = 10
    GUILD_BOOST_TIER_3 = 11
    CHANNEL_FOLLOW_ADD = 12
    GUILD_DISCOVERY_DISQUALIFIED = 14
    GUILD_DISCOVERY_REQUALIFIED = 15
    GUILD_DISCOVERY_GRACE_PERIOD_INITIAL_WARNING = 16
    GUILD_DISCOVERY_GRACE_PERIOD_FINAL_WARNING = 17
    THREAD_CREATED = 18
    REPLY = 19
    CHAT_INPUT_COMMAND = 20
    THREAD_STARTER_MESSAGE = 21
    GUILD_INVITE_REMINDER = 22
    CONTEXT_MENU_COMMAND = 23



from dataclasses import dataclass

from ...core.model import DataModel
from ...core.snowflake import Snowflake

@dataclass
class FollowedChannelModel(DataModel):
    """Represents the followed channel object."""

    channel_id: Snowflake
    """ID of the source channel."""

    webhook_id: Snowflake
    """Target webhook ID created."""



import logging

logger = logging.getLogger('scurrypy')

from scurrypy import Client
from scurrypy.bases import Addon
from scurrypy.enums import EventType
from scurrypy.core import DiscordError
from scurrypy.api.interactions import MessageComponentDataModel, ModalDataModel
from scurrypy.events import InteractionEvent

from .ctx import MessageComponentContext, ComponentModalContext

def _check_func_params(func: callable):
    import inspect
    
    params_len = len(inspect.signature(func).parameters)

    if params_len != 1:
        raise TypeError(
            f"Component handler '{func.__name__}' must accept exactly one parameter (ctx)."
        )

class ComponentsAddon(Addon):
    """Addon that implements automatic registering and decorating component interactions."""

    def __init__(self, client: Client):
        """
        Args:
            client (Client): the bot client object
        """
        self.bot = client

        self.component_handlers = {}
        """Mapping of component custom IDs to handler."""

        client.add_startup_hook(self.on_startup) # wait until start to register commands

    def on_startup(self):
        """Sets up the addon with the client."""

        self.bot.add_event_listener(EventType.INTERACTION_CREATE, self.dispatch)

    def component(self, custom_id: str, *, handler: callable = None):
        if handler is not None:
            _check_func_params(handler)
            self.component_handlers[custom_id] = handler
        else:
            def decorator(func):
                _check_func_params(func)
                self.component_handlers[custom_id] = func
            return decorator
    
    # helpers purly for ergonomics
    def button(self, custom_id: str, *, handler: callable = None):
        """Register and route button interactions.

        Args:
            custom_id (str): custom ID of button
                !!! warning "Important"
                    Must match the `custom_id` set where the component was created.
            handler (callable, optional): callback for the command (if not a decorator)
        """
        return self.component(custom_id, handler=handler)

    def select(self, custom_id: str, *, handler: callable = None):
        """Register and route select menu interactions.

        Args:
            custom_id (str): custom ID of select menu
                !!! warning "Important"
                    Must match the `custom_id` set where the component was created.
            handler (callable, optional): callback for the command (if not a decorator)
        """
        return self.component(custom_id, handler=handler)

    def modal(self, custom_id: str, *, handler: callable = None):
        """Register and route modal interactions.

        Args:
            custom_id (str): custom ID of modal
                !!! warning "Important"
                    Must match the `custom_id` set where the component was created.
            handler (callable, optional): callback for the command (if not a decorator)
        """
        return self.component(custom_id, handler=handler)

    def _get_handler(self, name: str):
        """Helper function for fetching a handler by `fnmatch`."""

        import fnmatch
        for k, v in self.component_handlers.items():
            if fnmatch.fnmatch(name, k):
                return v
        return False

    async def dispatch(self, event: InteractionEvent):
        """Dispatch a response to an `INTERACTION_CREATE` event

        Args:
            event (InteractionEvent): interaction event object
        """
        # only respond to component interactions
        data = event.data

        if not isinstance(data, (MessageComponentDataModel, ModalDataModel)):
            return # ignore non-component interactions

        name = data.custom_id
        handler = self._get_handler(name)

        if not handler:
            logger.warning(f"No handler registered for interaction '{name}'")
            return
                
        if isinstance(data, MessageComponentDataModel):
            ctx = MessageComponentContext(self.bot, event)
        
        elif isinstance(data, ModalDataModel):
            ctx = ComponentModalContext(self.bot, event)

        try:
            await handler(ctx)
            logger.info(f"Interaction '{name}' Acknowledged.")
        except DiscordError as e:
            logger.error(f"Error in interaction '{name}': {e}")
        except Exception as e:
            logger.exception(f"Unhandled error in interaction '{name}': {e}")



from typing import TypedDict, Optional

from ..api.commands.slash import CommandOptionPart

class EditGlobalCommandParams(TypedDict, total=False):
    """Parameters for editing a global command."""

    name: Optional[str]
    """Name of the command."""

    description: Optional[str]
    """Description for the command."""

    options: Optional[list[CommandOptionPart]]
    """Options with the command."""

    nsfw: Optional[bool]
    """Whether this command is age restricted."""

class EditGuildCommandParams(TypedDict, total=False):
    """Parameters for editing a guild command."""

    name: Optional[str]
    """Name of the command."""

    description: Optional[str]
    """Description for the command."""

    options: Optional[list[CommandOptionPart]]
    """Options with the command."""

    nsfw: Optional[bool]
    """Whether this command is age restricted."""



from dataclasses import dataclass

from ..core.model import DataModel
from ..core.snowflake import Snowflake

from .base_event import Event

from ..api.user import GuildMemberModel
from ..api.emoji import EmojiModel

@dataclass
class ReactionAddEvent(Event, DataModel):
    """Reaction added event."""

    type: int
    """Type of reaction added."""

    user_id: Snowflake
    """ID of user who added the emoji."""

    emoji: EmojiModel
    """Emoji used to react."""

    channel_id: Snowflake
    """ID of the channel where the reaction took place."""

    message_id: Snowflake
    """ID of the message where the reaction took place."""

    guild_id: Snowflake | None
    """ID of the guild where the reaction took place (if in a guild)."""

    burst: bool
    """Whether the emoji is super."""

    member: GuildMemberModel | None
    """Partial member object of the guild member that added the emoji (if in a guild)."""

    message_author_id: Snowflake | None
    """ID of the user who sent the message where the reaction was added."""

@dataclass
class ReactionRemoveEvent(Event, DataModel):
    """Reaction removed event."""

    type: int
    """Type of reaction removed."""

    user_id: Snowflake
    """ID of user who removed their reaction."""

    emoji: EmojiModel
    """Emoji data of the emoji where the reaction was removed."""

    channel_id: Snowflake
    """ID of the channel where the reaction was removed."""

    message_id: Snowflake
    """ID of the message where the reaction was removed."""

    guild_id: Snowflake | None
    """ID of the guild where the reaction was removed (if in a guild)."""

    burst: bool
    """If the emoji of the removed reaction is super."""

@dataclass
class ReactionRemoveEmojiEvent(Event, DataModel):
    """All reactions of a specific emoji removed."""

    emoji: EmojiModel
    """Emoji data of the removed reaction emoji."""

    channel_id: Snowflake
    """ID of the channel where the reaction emoji was removed."""

    message_id: Snowflake
    """ID of the message where the reaction emoji was removed."""

    guild_id: Snowflake | None
    """ID of the guild where the reaction emoji was removed. (if in a guild)"""

@dataclass
class ReactionRemoveAllEvent(Event, DataModel):
    """Remove all reactions event."""

    channel_id: Snowflake
    """ID of the channel where all reaction were removed."""

    message_id: Snowflake
    """ID of the message where all reaction were removed."""

    guild_id: Snowflake | None
    """ID of the guild where all reaction were removed (if in a guild)."""



from scurrypy import Client
from scurrypy.resources import Interaction
from scurrypy.events import InteractionEvent

class InteractionContext(Interaction):
    """Useful interaction event info."""

    def __init__(self, bot: Client, event: InteractionEvent):
        super().__init__(bot.http, event.id, event.token)
        self.bot = bot
        self.event = event
        self.data = event.data

    @property
    def user(self):
        """The invoking user."""
        return self.event.member.user
    
    @property
    def channel(self):
        """Channel resource of the interaction."""
        return self.bot.channel(self.event.channel_id)
    
    @property
    def guild(self):
        """Guild resource of the interaction."""
        return self.bot.guild(self.event.guild_id)

    @property
    def message(self):
        """Message resource of the interaction."""
        return self.bot.message(self.event.channel_id, self.event.message.id)



from dataclasses import dataclass

from ..core.model import DataModel
from ..core.snowflake import Snowflake

from .base_event import Event

from ..api.channels.channel import ChannelModel
from ..api.guilds.guild import UnavailableGuildModel, GuildModel
from ..api.messages.sticker import StickerModel

from ..api.emoji import EmojiModel
from ..api.user import UserModel, GuildMemberModel

@dataclass
class GuildCreateEvent(Event, GuildModel):
    """Received when the bot has joined a guild."""
    
    joined_at: str
    """ISO8601 timestamp of when app joined the guild."""

    large: bool
    """If the guild is considered large."""

    member_count: int
    """Total number of members in the guild."""

    members: list[GuildMemberModel]
    """Users in the guild."""

    channels: list[ChannelModel]
    """Channels in the guild."""

    threads: list[ChannelModel]
    """All active threads in the guild that are viewable."""

    unavailable: bool | None
    """`True` if the guild is unavailable due to an outage."""

@dataclass
class GuildUpdateEvent(Event, DataModel):
    """Received when a guild has been edited."""

    id: Snowflake
    """ID of the guild."""

    name: str
    """Name of the guild."""

    icon: str
    """Image hash of the guild's icon."""

    description: str
    """Description of the guild."""

    banner: str
    """Image hash of the guild's banner."""

    joined_at: str
    """ISO8601 timestamp of when app joined the guild."""

    large: bool
    """If the guild is considered large."""

    member_count: int
    """Total number of members in the guild."""

@dataclass
class GuildDeleteEvent(Event, UnavailableGuildModel):
    """Received when the bot has left a guild or the guild was deleted."""
    pass

@dataclass
class GuildBanAddEvent(Event, DataModel):
    """Received when a user is banned from a guild."""

    guild_id: Snowflake
    """ID of the guild in which the ban took place."""

    user: UserModel
    """The user who was banned."""

@dataclass
class GuildBanRemoveEvent(Event, DataModel):
    """Received when a user is unbanned from a guild."""

    guild_id: Snowflake
    """ID of the guild in which the ban took place."""

    user: UserModel
    """The user who was banned."""

@dataclass
class GuildEmojisUpdateEvent(Event, DataModel):
    """Received when a guild updates their emojis."""

    guild_id: Snowflake
    """ID of the guild."""

    emojis: list[EmojiModel]
    """Complete set of guild emojis with changes."""

@dataclass
class GuildStickersUpdateEvent(Event, DataModel):
    """Received when a guild's stickers have been updated."""

    guild_id: Snowflake
    """ID of the guild."""

    stickers: list[StickerModel]
    """List of the guild's stickers."""



# scurrypy/api/channels

from .announcement import GuildAnnouncementChannelPart
from .channel import ChannelModel
from .default_reaction import (
    DefaultReactionModel, 
    DefaultReactionPart
)
from .dm import DMChannelModel
from .followed import FollowedChannelModel
from .forum import (
    GuildForumChannelModel, 
    GuildForumChannelPart
)
from .guild_text import GuildTextChannelPart
from .tag import (
    TagModel, 
    TagPart
)
from .threads import (
    ThreadFromMessagePart, 
    ThreadWithoutMessagePart, 
    ThreadMetadataModel, 
    ThreadMemberModel, 
    ArchivedThreadsModel, 
    ActiveThreadsModel, 
    ThreadChannelModel
)

__all__ = [
    "GuildAnnouncementChannelPart",

    "ChannelModel",

    "DefaultReactionModel", 
    "DefaultReactionPart",
    
    "DMChannelModel",

    "FollowedChannelModel",

    "GuildForumChannelModel", 
    "GuildForumChannelPart",

    "GuildTextChannelPart",

    "TagModel", 
    "TagPart",

    "ThreadFromMessagePart", 
    "ThreadWithoutMessagePart", 
    "ThreadMetadataModel", 
    "ThreadMemberModel", 
    "ArchivedThreadsModel", 
    "ActiveThreadsModel", 
    "ThreadChannelModel"
]



from scurrypy.api.interactions import (
    ApplicationCommandDataModel, 
    AutocompleteApplicationCommandDataModel,
)

from ..interactions.ctx import InteractionContext

class ApplicationCommandContext(InteractionContext):
    data: ApplicationCommandDataModel

class AutocompleteApplicationCommandContext(InteractionContext):
    data: AutocompleteApplicationCommandDataModel



from .enum_types import DiscordTypes

class CommandType(DiscordTypes):
    """Types of commands."""
    
    CHAT_INPUT = 1
    """Slash commands; a text-based command that shows up when a user types `/`."""

    USER = 2
    """A UI-based command that shows up when you right click or tap on a user."""
    
    MESSAGE = 3
    """A UI-based command that shows up when you right click or tap on a message."""

class CommandOptionType(DiscordTypes):
    """Slash command option input types."""

    STRING = 3
    """string (text)"""

    INTEGER = 4
    """integer (Any integer between -2^53+1 and 2^53-1)"""

    BOOLEAN = 5
    """boolean (true/false)"""

    USER = 6
    """user pangination"""

    CHANNEL = 7
    """channel pangination (category and channels)"""

    ROLE = 8
    """role pangination"""

    MENTIONABLE = 9
    """any pangination (role and user)"""

    NUMBER = 10
    """number (Any double between -2^53 and 2^53)"""

    ATTACHMENT = 11
    """File upload. See [`AttachmentPart`][scurrypy.api.messages.AttachmentPart]."""



# scurrypy/params

from .emoji import EditApplicationEmojiParams
from .channel import (
    EditGuildChannelParams, 
    EditThreadChannelParams
)
from .command import (
    EditGuildCommandParams, 
    EditGlobalCommandParams
)
from .emoji import EditGuildEmojiParams
from .guild import (
    EditGuildRoleParams, 
    EditGuildParams, 
    EditGuildWelcomeScreenParams, 
    EditOnboardingParams,
    EditGuildStickerParams
)
from .message import EditMessageParams
from .user import (
    EditGuildMemberParams, 
    EditUserParams
)

__all__ = [
    "EditApplicationEmojiParams",

    "EditGuildChannelParams", 
    "EditThreadChannelParams",

    "EditGuildCommandParams", 
    "EditGlobalCommandParams",

    "EditGuildEmojiParams",

    "EditGuildRoleParams", 
    "EditGuildParams", 
    "EditGuildWelcomeScreenParams", 
    "EditOnboardingParams",

    "EditGuildStickerParams",

    "EditMessageParams",

    "EditGuildMemberParams", 
    "EditUserParams"
]



from dataclasses import dataclass

from ..core.model import DataModel
from ..core.snowflake import Snowflake

from .base_event import Event

from ..api.guilds.role import GuildRoleModel

@dataclass
class RoleCreateEvent(Event, DataModel):
    """Received when a guild role is created."""

    guild_id: Snowflake
    """Guild ID of the role."""

    role: GuildRoleModel
    """The new role."""

@dataclass
class RoleUpdateEvent(Event, DataModel):
    """Received when a guild role is updated."""

    guild_id: Snowflake
    """Guild ID of the role."""

    role: GuildRoleModel
    """The new role."""

@dataclass
class RoleDeleteEvent(Event, DataModel):
    """Received when a guild role is deleted."""

    guild_id: Snowflake
    """Guild ID of the role."""

    role_id: Snowflake
    """Role ID of the role."""



from .enum_types import DiscordTypes

class InteractionDataType(DiscordTypes):
    """Interaction data types constants."""

    SLASH_COMMAND = 1
    """The interaction is a slash command."""

    USER_COMMAND = 2
    """The interaction is attached to a user."""

    MESSAGE_COMMAND = 3
    """The interaction is attached to a message."""

class InteractionType(DiscordTypes):
    """Interaction types constants."""

    APPLICATION_COMMAND = 2
    """Slash command interaction."""

    MESSAGE_COMPONENT = 3
    """Message component interaction (e.g., button, select menu, etc.)."""

    APPLICATION_COMMAND_AUTOCOMPLETE = 4
    """Application command autocompletion."""

    MODAL_SUBMIT = 5
    """Modal submit interaction."""

class InteractionCallbackType(DiscordTypes):
    """Interaction callback types constants."""

    PONG = 1
    """Acknowledge a Ping."""

    CHANNEL_MESSAGE_WITH_SOURCE = 4
    """Respond to an interaction with a message."""

    DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE = 5
    """Acknowledge an interaction and edit a response later. User sees a loading state."""

    DEFERRED_UPDATE_MESSAGE = 6
    """
        Acknowledge an interaction and edit the original message later. 
        The user does NOT see a loading state. (Components only)
    """

    UPDATE_MESSAGE = 7
    """Edit the message in which the component was attached."""

    APPLICATION_COMMAND_AUTOCOMPLETE_RESULT = 8
    """Respond to an autocomplete interaction with suggested choices."""

    MODAL = 9
    """Respond to an interaction with a popup modal (not available for MODAL_SUBMIT and PING interactions)."""



# scurrypy/api/interactions

from .interaction import (
    InteractionCallbackDataModel, 
    InteractionCallbackModel, 
    InteractionModel, 
    CommandDataModel,
    ApplicationCommandOptionDataModel, 
    ApplicationCommandDataModel, 
    MessageComponentDataModel, 
    ModalComponentDataModel, 
    ModalComponentModel, 
    ModalDataModel,
    AutocompleteApplicationCommandDataModel
)
from .modal import ModalPart
from .resolved import ResolvedDataModel

__all__ = [
    "InteractionCallbackDataModel", 
    "InteractionCallbackModel", 
    "InteractionModel", 
    "CommandDataModel",
    "ApplicationCommandOptionDataModel", 
    "ApplicationCommandDataModel", 
    "MessageComponentDataModel", 
    "ModalComponentDataModel", 
    "ModalComponentModel", 
    "ModalDataModel",

    "ModalPart",
    "ResolvedDataModel",
    "AutocompleteApplicationCommandDataModel"
]



from dataclasses import dataclass

from ..core.model import DataModel
from ..core.snowflake import Snowflake

from ..enums.invite import InviteType

from .guilds.guild import GuildModel
from .guilds.role import GuildRoleModel
from .channels.channel import ChannelModel

from .user import UserModel

@dataclass
class InviteModel(DataModel):
    """Represents a code that adds a user to guild or group DM channel."""

    type: InviteType
    """Type of invite."""

    code: str
    """Invite code (unique ID)."""

    guild: GuildModel | None
    """Guild the invite is for."""

    channel: ChannelModel
    """Channel this invite is for."""

    inviter: UserModel | None
    """User who created invite."""

    approximate_member_count: int | None
    """Approximate count of total members."""

    expires_at: str
    """ISO8601 timestamp for expiration date."""

    roles: list[GuildRoleModel] | None
    """Roles assigned to the user upon accepting the invite."""

@dataclass
class InviteWithMetadataModel(InviteModel):
    """Represents the invite model with extra information."""

    uses: int
    """Number of times this invite was used."""

    max_uses: int
    """Max number of times this invite can be used."""

    max_age: int
    """Duration (in seconds) after which this invite expires."""

    temporary: bool
    """Whether this invite only grants temporary membership."""

    created_at: str
    """ISO8601 timestamp for when this invite was created."""

@dataclass
class InvitePart(DataModel):
    """Represents fields for creating an invite."""

    max_age: int | None = None
    """Duration of invite (in seconds) before it expires. 
    `0` for never or up to `604800` (max 7 days).
    Discord defaults to `86400` (24 hours).
    """

    max_uses: int | None = None
    """Max number of uses for this invite.
    `0` for unlimited or up to `100`.
    Discord defaults to `0`.
    """

    temporary: bool | None = None
    """Whether this invite grants temporary membership.
    Discord defaults to `False`.
    """

    unique: bool | None = None
    """Whether to reuse similar invite codes.
    Discord defaults to `False`.
    """

    role_ids: list[Snowflake] | None = None
    """Role IDs to be given when the user accept this invite.
    
    !!! important "Permissions"
        Requires `MANAGE_ROLES` and cannot assign roles with higher
        permissions than the sender.
    """



from dataclasses import dataclass

from ...core.model import DataModel
from ...core.snowflake import Snowflake

from ...bases.interaction import InteractionData

from ...enums.permissions import Permissions
from ...enums.interaction import InteractionCallbackType, InteractionType
from ...enums.command import CommandType, CommandOptionType
from ...enums.components import ComponentType

from ..channels.channel import ChannelModel
from ..guilds.guild import GuildModel
from ..messages.message import MessageModel

from ..user import GuildMemberModel

from .resolved import ResolvedDataModel

@dataclass
class InteractionCallbackDataModel(DataModel):
    """Represents the interaction callback object."""

    id: Snowflake
    """ID of the interaction."""

    type: InteractionCallbackType
    """Type of interaction."""

    activity_instance_id: str
    """Instance ID of activity if an activity was launched or joined."""

    response_message_id: Snowflake
    """ID of the message created by the interaction."""

    response_message_loading: bool
    """If the interaction is in a loading state."""

    response_message_ephemeral: bool
    """If the interaction is ephemeral."""

@dataclass
class InteractionCallbackModel(DataModel):
    """Represents the interaction callback response object."""

    interaction: InteractionCallbackDataModel
    """The interaction object associated with the interaction response."""

@dataclass
class InteractionModel(DataModel):
    """Represents the interaction model."""

    type: InteractionType
    """Type of interaction."""

    id: Snowflake
    """ID of interaction."""

    token: str
    """token of interaction."""

    channel_id: Snowflake
    """ID of the channel where the interaction was sent."""

    application_id: Snowflake
    """ID of the application that owns the interaction."""

    app_permissions: Permissions
    """Bitwise set of permissions pertaining to the location of the interaction. [`INT_LIMIT`]"""

    member: GuildMemberModel
    """Guild member invoking the interaction."""

    message: MessageModel | None
    """Message associated with interaction (components or modals)."""

    guild_id: Snowflake | None
    """ID of guild the interaction was invoked (if invoked in a guild)."""

    guild: GuildModel | None
    """Partial guild object of the guild the interaction was invoked (if invoked in a guild)."""

    channel: ChannelModel | None
    """Partial channel object the interaction was invoked."""

# ----- Command Interaction -----

@dataclass
class ApplicationCommandOptionDataModel(DataModel):
    """Represents the response options from a slash command."""
    
    name: str
    """Name of the command option."""

    type: CommandOptionType
    """Type of command option."""

    value: str
    """
    Raw value from Discord as a string.
    
    Convert based on option type:
        - INTEGER/USER/CHANNEL/ROLE/ATTACHMENT: int(value)
        - NUMBER: float(value)  
        - BOOLEAN: value.lower() == 'true'

    otherwise the value is expected to be str
    """

    focused: bool
    """Whether this option is the currently focused option for autocomplete."""

@dataclass
class CommandDataModel(DataModel, InteractionData):
    """Represents common command interaction data fields."""
    
    id: Snowflake
    """ID of the command."""

    name: str
    """Name of the command."""
    
    type: CommandType
    """Type of command (e.g., message, user, slash)."""

    guild_id: Snowflake | None
    """ID of guild from which the command was invoked."""

    target_id: Snowflake  | None
    """ID of the user or message from which the command was invoked (message/user commands only)."""

    resolved: ResolvedDataModel | None
    """Converted users + roles + channels + attachments."""

    options: list[ApplicationCommandOptionDataModel] | None
    """Options of the command (slash command only)."""

    def get_focused_value(self):
        opt = next((opt for opt in self.options if opt), None)
        
        return opt.value if opt else ""

    def get_option(self, option_name: str, default = None):
        """Get the input for a command option by name and convert it to its proper type.

        Args:
            option_name (str): option to fetch input from

        Returns:
            (int | float | bool | str | Any): converted input data of specified option
        """
        for option in self.options:
            if option.name != option_name:
                continue

            if option.type in [
                CommandOptionType.INTEGER,
                CommandOptionType.USER,
                CommandOptionType.CHANNEL,
                CommandOptionType.ROLE,
                CommandOptionType.ATTACHMENT
            ]:
                return int(option.value)
            
            if option.type == CommandOptionType.NUMBER:
                return float(option.value)
            
            if option.type == CommandOptionType.BOOLEAN:
                return option.value.lower() == 'true'
            
            return option.value
        
        return default

@dataclass
class ApplicationCommandDataModel(CommandDataModel):
    """Represents the response from a command."""
    pass


@dataclass
class AutocompleteApplicationCommandDataModel(CommandDataModel):
    """Represents the response from an autocomplete command."""
    pass

# ----- Component Interaction -----

@dataclass
class MessageComponentDataModel(DataModel, InteractionData):
    """Represents the select response from a select component."""

    custom_id: str
    """Unique ID associated with the component."""

    component_type: int
    """Type of component."""

    resolved: ResolvedDataModel | None
    """Resolved entities from selected options."""

    values: list[str] | None
    """Select values (if any)."""

# ----- Modal Interaction -----

@dataclass
class ModalComponentDataModel(DataModel):
    """Represents the modal field response from a modal."""

    type: ComponentType
    """Type of component."""
    
    value: str | None
    """Text input value.
    
    Convert based on option type:
        - CHECKBOX: value.lower() == 'true'

    otherwise the value is expected to be str
    """

    custom_id: str
    """Unique ID associated with the component."""

    values: list[str] | None
    """String select values."""

@dataclass
class ModalComponentModel(DataModel):
    """Represents the modal component response from a modal."""

    type: ComponentType
    """Type of component."""

    component: ModalComponentDataModel
    """Data associated with the component."""

@dataclass
class ModalDataModel(DataModel, InteractionData):
    """Represents the modal response from a modal."""
    
    custom_id: str
    """Unique ID associated with the modal."""

    resolved: ResolvedDataModel | None
    """Resolved entities from modal data."""

    components: list[ModalComponentModel]
    """Components on the modal."""

    def get_modal_data(self, custom_id: str):
        """Fetch a modal field's data by its custom ID

        Args:
            custom_id (str): custom ID of field to fetch

        Raises:
            (ValueError): invalid custom ID

        Returns:
            (str, list[str], bool): component values (if select component) or value or bool if checkbox
        """
        from ...enums.components import ComponentType
        
        for component in self.components:
            if custom_id != component.component.custom_id:
                continue

            t = component.component.type

            if t in [
                ComponentType.STRING_SELECT, 
                ComponentType.USER_SELECT, 
                ComponentType.ROLE_SELECT, 
                ComponentType.MENTIONABLE_SELECT, 
                ComponentType.CHANNEL_SELECT,
                ComponentType.FILE_UPLOAD,
                ComponentType.CHECKBOX_GROUP
            ]:
                return component.component.values
            
            if t == ComponentType.CHECKBOX: # value is bool for checkboxes
                return component.component.value.lower() == 'true'
            
            # text input
            return component.component.value

        raise ValueError(f"Component custom ID '{custom_id}' not found.")



from dataclasses import dataclass

from .base_resource import BaseResource

from ..api.invite import InviteModel

@dataclass
class Invite(BaseResource):
    """Represents a Discord invite."""
    
    code: str
    """Invite code."""

    async def fetch(self, with_counts: bool = None) -> InviteModel:
        """Fetch the invite object for the given code.

        Args:
            with_counts (bool, optional): whether the model should contain approximate member counts

        Returns:
            (InviteModel): queried invite object
        """
        data = await self.http.request(
            'GET', 
            f'/invites/{self.code}', 
            params={'with_counts': with_counts}
        )

        return InviteModel.from_dict(data)

    async def delete(self) -> None:
        """Delete the invite for the given code.
        Fires [`InviteDeleteEvent`][scurrypy.events.invite_events.InviteDeleteEvent].
        
        !!! important "Permissions"
            Requires `MANAGE_CHANNELS` on the channel this invite belongs to
            or `MANAGE_GUILD` to remove any invite across the guild

        Returns:
            (InviteModel): deleted invite object
        """
        data = await self.http.request('DELETE', f'/invites/{self.code}')

        return InviteModel.from_dict(data)



from dataclasses import dataclass

from ..core.model import DataModel
from ..core.snowflake import Snowflake

from .base_event import Event

from ..api.integration import IntegrationModel

@dataclass
class GuildIntegrationCreateEvent(Event, IntegrationModel):
    """Received when an integration is created."""

    guild_id: Snowflake
    """Guild ID of the created integration."""

@dataclass
class GuildIntegrationUpdateEvent(Event, IntegrationModel):
    """Received when an integration is created."""

    guild_id: Snowflake
    """Guild ID of the updated integration."""

@dataclass
class GuildIntegrationsUpdateEvent(Event, DataModel):
    """Received when a guild's integration is updated."""

    guild_id: Snowflake
    """ID of the guild whose integrations were updated."""

@dataclass
class GuildIntegrationDeleteEvent(Event, DataModel):
    """Received when a guild's integration is deleted."""

    id: Snowflake
    """ID of the deleted integration."""

    guild_id: Snowflake
    """Guild ID of the deleted integration."""

    application_id: Snowflake | None
    """ID of the bot for this Discord integration."""



from scurrypy import Client
from scurrypy.bases import Addon
from scurrypy.enums import EventType
from scurrypy.core import Snowflake, DiscordError
from scurrypy.api.guilds import GuildRoleModel
from scurrypy.events import GuildCreateEvent, GuildDeleteEvent, RoleCreateEvent, RoleUpdateEvent, RoleDeleteEvent

class GuildRoleCacheAddon(Addon):
    """Defines caching guild roles and lookup."""

    def __init__(self, client: Client):
        self.bot = client

        self.roles: dict[Snowflake, dict[Snowflake, GuildRoleModel]] = {} # stores OBJECTS
        self.role_index: dict[Snowflake, GuildRoleModel] = {} # stores REFERENCES

        client.add_event_listener(EventType.GUILD_CREATE, self.on_guild_create)
        client.add_event_listener(EventType.GUILD_DELETE, self.on_guild_delete)

        client.add_event_listener(EventType.ROLE_CREATE, self.on_role_create)
        client.add_event_listener(EventType.ROLE_UPDATE, self.on_role_update)
        client.add_event_listener(EventType.ROLE_DELETE, self.on_role_delete)

    def on_guild_create(self, event: GuildCreateEvent):
        """Append new guild roles to cache. Also add roles to index.

        Args:
            event (GuildCreateEvent): the GUILD_CREATE event
        """
        guild_dict = self.roles.setdefault(event.id, {})

        for role in event.roles:
            guild_dict[role.id] = role
            self.role_index[role.id] = role

    def on_guild_delete(self, event: GuildDeleteEvent):
        """Remove guild roles from cache. Also remove roles from index

        Args:
            event (GuildDeleteEvent): the GUILD_DELETE event
        """
        removed_roles = self.roles.pop(event.id, {})

        for role in removed_roles.values():
            self.role_index.pop(role.id, None)

    def on_role_create(self, event: RoleCreateEvent):
        """Append role to guild key. Also append role to index.

        Args:
            event (RoleCreateEvent): the ROLE_CREATE event
        """
        model = GuildRoleModel.from_dict(event.raw)
        guild_dict = self.roles.setdefault(event.guild_id, {})

        guild_dict[event.role.id] = model
        self.role_index[event.role.id] = model

    def on_role_update(self, event: RoleUpdateEvent):
        """Replace role in guild key. Also replace role in index.

        Args:
            event (RoleUpdateEvent): the ROLE_UPDATE event
        """
        model = GuildRoleModel.from_dict(event.raw)
        guild_dict = self.roles.setdefault(event.guild_id, {})

        guild_dict[event.role.id] = model
        self.role_index[event.role.id] = model

    def on_role_delete(self, event: RoleDeleteEvent):
        """Remove role from guild key. Also remove role from index.

        Args:
            event (RoleDeleteEvent): the ROLE_DELETE event
        """
        model = self.role_index.pop(event.role_id, None)
        if model:
            self.roles.get(event.guild_id, {}).pop(event.role_id, None)

    def get_role(self, role_id: Snowflake):
        """Get a role from the cache.

        Args:
            role_id (Snowflake): ID of the role

        Returns:
            (GuildRoleModel | None): the role object if found else None
        """
        return self.role_index.get(role_id)
    
    async def get_role(self, guild_id: Snowflake, role_id: Snowflake):
        """Fetch a guild role. If not found, request and store it.

        Args:
            guild_id (Snowflake): guild ID of role
            role_id (Snowflake): role ID of guild

        Returns:
            (GuildRoleModel | None): hydrated role object or None if fetch failed
        """
        role = self.role_index.get(role_id)
        if role:
            return role
        
        try:
            role = await self.bot.guild(guild_id).fetch_guild_role(role_id)
        except DiscordError:
            return None
        
        self.put(guild_id, role)
        return role

    def put(self, guild_id: Snowflake, role: GuildRoleModel):
        """Put a new role into the cache.

        Args:
            guild_id (Snowflake): guild ID of the role
            role (GuildRoleModel): the role object
        """
        guild_dict = self.roles.setdefault(guild_id, {})
        guild_dict[role.id] = role
        self.role_index[role.id] = role



from typing import TypedDict

from ..api.image_data import ImageDataPart

class EditGuildMemberParams(TypedDict, total=False):
    """Parameters for editing a guild member."""

    nick: str
    """User's guild nickname.
    
    !!! important "Permissions"
        Requires `MANAGE_NICKNAMES`
    """

    roles: list[int]
    """Role IDs the member is assigned.
    
    !!! important "Permissions"
        Requires `MANAGE_ROLES`
    """

class EditUserParams(TypedDict, total=False):
    """Parameters for editing a user."""

    username: str
    """User's username.
    
    !!! note
        May cause the discriminator to be randomized.
    """

    avatar: ImageDataPart
    """User's avatar."""

    banner: ImageDataPart
    """User's banner."""



from typing import TypedDict

class EditApplicationEmojiParams(TypedDict):
    """Represents options for editing a bot emoji."""
    
    name: str
    """Name of the emoji."""

class EditGuildEmojiParams(TypedDict, total=False):
    """Parameters for editing a guild emoji."""

    name: str
    """Name of the moji."""

    roles: list[int]
    """Roles allowed to use this emoji."""



from dataclasses import dataclass, field

from ...core.model import DataModel
from ...core.snowflake import Snowflake

from ...bases.components import (
    ActionRowChild, 
    LabelChild,
    Component
)

from ...enums.components import ComponentType, DefaultValueType

from ..emoji import EmojiModel

@dataclass
class SelectOption(DataModel):
    """Represents the Select Option component."""

    label: str | None = None
    """User-facing name of the option."""

    value: str | None = None
    """Developer-defined value of the option."""

    description: str | None = None
    """Additional description of the option."""

    emoji: EmojiModel | None = None
    """Partial emoji object."""

    default: bool | None = None
    """Whether this option is selected by default. Discord defaults to `False`."""

@dataclass
class StringSelect(DataModel, Component, ActionRowChild, LabelChild):
    """Represents the String Select component.
    
    A String Select allows users to select one or more provided options.
    """

    custom_id: str | None = None
    """ID for the select menu."""

    options: list[SelectOption] | None = None
    """Specified choices in a select menu."""

    placeholder: str | None = None
    """Placeholder text if nothing is selected or default."""

    min_values: int | None = None
    """Minimum number of items that must be chosen. Discord defaults to `1`."""

    max_values: int | None = None
    """Maximum number of items that can be chosen. Discord defaults to `1`."""

    required: bool | None = None
    """Whether the string select is required to answer in a modal. Discord defaults to `True`."""

    disabled: bool | None = None # does not work on Modals!
    """Whether select menu is disabled in a message. Discord defaults to `False`."""

    type: ComponentType = field(init=False, default=ComponentType.STRING_SELECT)
    """Component type. Always `ComponentType.STRING_SELECT` for this class."""


@dataclass
class DefaultValue(DataModel):
    """Represents the Default Value for Select components."""

    id: Snowflake | None = None
    """ID of role, user, or channel."""

    type: DefaultValueType | None = None
    """Type of value that `id` represents."""

@dataclass
class SelectMenu(DataModel):
    """Represents common fields for Discord's select menus."""

    custom_id: str = None
    """ID for the select menu."""

    placeholder: str | None = None
    """Placeholder text if nothing is selected."""

    default_values: list[DefaultValue] | None = None
    """
        List of default values for auto-populated select menu components.
        Number of default values must be in the range of `min_values` to `max_values`.
    """

    min_values: int | None = None
    """Minimum number of items that must be chosen. Discord defaults to `1`."""

    max_values: int | None = None
    """Maximum number of items that can be chosen. Discord defaults to `1`."""

    required: bool | None = None
    """Whether the select is required to answer in a modal. Discord defaults to `True`."""

    disabled: bool | None = None
    """Whether select menu is disabled in a message. Discord defaults to `False`."""

@dataclass
class UserSelect(SelectMenu, Component, ActionRowChild, LabelChild):
    """Represents the User Select component.
    
    User Select allows users to select one or more users.
    """

    type: ComponentType = field(init=False, default=ComponentType.USER_SELECT)
    """Component type. Always `ComponentType.USER_SELECT` for this class."""

@dataclass
class RoleSelect(SelectMenu, Component, ActionRowChild, LabelChild):
    """Represents the Role Select component.
    
    A Role Select allows users to select one or more roles
    """

    type: ComponentType = field(init=False, default=ComponentType.ROLE_SELECT)
    """Component type. Always `ComponentType.ROLE_SELECT` for this class."""

@dataclass
class MentionableSelect(SelectMenu, Component, ActionRowChild, LabelChild):
    """Represents the Mentionable Select component.
    
    A Mentionable Select allows users to select one or more mentionables.
    """

    type: ComponentType = field(init=False, default=ComponentType.MENTIONABLE_SELECT)
    """Component type. Always `ComponentType.MENTIONABLE_SELECT` for this class."""

@dataclass
class ChannelSelect(SelectMenu, Component, ActionRowChild, LabelChild):
    """Represents the Channel Select component.
    
    A Channel Select allows users to select one or more channels.
    """

    type: ComponentType = field(init=False, default=ComponentType.CHANNEL_SELECT)
    """Component type. Always `ComponentType.CHANNEL_SELECT` for this class."""



from dataclasses import dataclass

from ...core.model import DataModel
from ...core.snowflake import Snowflake

from ...enums.guild import StickerType, StickerFormatType

from ..user import UserModel

@dataclass
class StickerModel(DataModel):
    """Represents the sticker object."""

    id: Snowflake
    """ID of the sticker."""

    pack_id: Snowflake | None
    """ID of the pack the sticker is from (if standard)."""

    name: str
    """Name of the sticker."""

    description: str
    """Description of the sticker."""

    tags: str
    """Autocomplete/suggestion tags for the sticker."""

    type: StickerType
    """Type of sticker."""

    format_type: StickerFormatType
    """Type of sticker format."""

    available: bool | None
    """Whether this guild sticker can be used.
    
    !!! note
        May be `False` due to loss of Server Boosts
    """

    guild_id: Snowflake | None
    """ID of the guild that owns this sticker."""
    
    user: UserModel | None
    """The user that uploaded the guild sticker."""

    sort_type: int | None
    """The standard sticker's sort order within its pack."""

@dataclass
class StickerItemModel(DataModel):
    """Represents a minimal sticker item."""
    
    id: Snowflake
    """ID of the sticker."""

    name: str
    """Name of the sticker."""

    format_type: StickerFormatType
    """Type of sticker format."""

@dataclass
class StickerPackModel(DataModel):
    """Represents a pack of standard stickers."""

    id: Snowflake
    """ID of the sticker pack."""

    stickers: list[StickerModel]
    """The stickers in the pack."""

    name: str
    """Name of the sticker pack."""

    sku_id: Snowflake
    """ID of the pack's SKU."""

    cover_sticker_id: Snowflake | None
    """ID of a sticker in the pack which is shown as the pack's icon."""

    description: str
    """Description of the sticker pack."""

    banner_asset_id: Snowflake | None
    """ID of the sticker pack's banner image."""

@dataclass
class StickerPart(DataModel):
    """Represents fields for creating a sticker."""

    name: str | None = None
    """Name of the sticker."""

    description: str | None = None
    """Description of the sticker."""

    tags: str | None = None
    """Autocomplete/suggestion tags for the sticker."""



import logging

logger = logging.getLogger('scurrypy')

from scurrypy import Client
from scurrypy.bases import Addon
from scurrypy.enums import EventType
from scurrypy.core import DiscordError
from scurrypy.events import Event

def _check_func_params(func: callable):
    import inspect
    
    params_len = len(inspect.signature(func).parameters)

    if params_len != 2:
        raise TypeError(f"Event handler '{func.__name__}' must accept exactly two parameters (bot, event).")

class EventsAddon(Addon):
    """Addon that implements automatic registering and decorating events."""

    def __init__(self, client: Client):
        """
        Args:
            client (Client): the Client object
        """
        self.bot = client

        self._events: dict[str, list[tuple[int, callable]]] = {}
        """Maps EVENT_NAME to handlers."""

        client.add_startup_hook(self.on_startup)

    def on_startup(self):
        """Adds registered events to client's event listener."""

        # lead all registered events to this dispatch
        for dispatch_type in self._events.keys():
            self.bot.add_event_listener(dispatch_type, self.dispatch)
        
    def listen(self, event_name: EventType, *, handler: callable = None):
        """Register and route an event with params (bot, event).

        Args:
            event_name (str): event name
            handler (callable, optional): callback for the event (if not a decorator)
        """
        if handler is not None:
            _check_func_params(handler)
            self._events.setdefault(event_name, []).append(handler)
        else:
            def decorator(func):
                _check_func_params(func)
                self._events.setdefault(event_name, []).append(func)
            return decorator

    async def dispatch(self, event: Event):
        """Addon's entry point.

        Args:
            event (Event): event data object
        """
        handlers = self._events.get(event.name)

        if not handlers:
            return
    
        try:
            for handler in handlers:
                await handler(self.bot, event)
        except DiscordError as e:
            logger.error(f"Error in event '{handler}': {e}")
        except Exception as e:
            logger.exception(f"Unhandled error in event '{handler.__name__}': {e}")



from dataclasses import dataclass

from ...core.model import DataModel
from ...core.snowflake import Snowflake

from ...enums.permissions import Permissions
from ...enums.channel import ChannelFlags

@dataclass
class ChannelModel(DataModel):
    """Represents common channel fields."""

    id: Snowflake
    """ID of the channel."""

    guild_id: Snowflake | None
    """Guild ID of the channel."""

    parent_id: Snowflake | None
    """Category ID of the channel."""

    position: int | None
    """Position of the channel."""

    name: str | None
    """Name of the channel."""

    topic: str | None
    """Topic of the channel."""

    nsfw: bool | None
    """If the channel is flagged NSFW."""

    last_message_id: Snowflake | None
    """ID of the last message sent in the channel."""

    rate_limit_per_user: int | None
    """Seconds user must wait between sending messages in the channel."""

    last_pin_timestamp: str | None
    """ISO8601 timestamp of the last pinned messsage in the channel."""

    permissions: Permissions | None
    """Permissions for the invoking user in this channel.
        Includes role and overwrite calculations. [`INT_LIMIT`]
    """

    flags: ChannelFlags
    """Channel flags combined as a bitfield."""



from typing import get_args, get_origin, Union
from types import UnionType

from ..bases.components import Component

def serialize(val) -> dict:
    """Serialize the value.

    Args:
        val (Any): value to serialize

    Returns:
        (dict): serialized value
    """
    if hasattr(val, "to_dict"):
        return val.to_dict()
    
    if isinstance(val, list):
        return [serialize(v) for v in val if v is not None]
    
    if isinstance(val, dict):
        return {k: serialize(v) for k, v in val.items()}

    return val

def convert(t, v):
    """Convert the given value to the given type.

    Args:
        t (type): type in which to convert value
        v (Any): value to be converted

    Raises:
        (TypeError): ambiguous type

    Returns:
        (type): converted value
    """
    o = get_origin(t)

    if o in (Union, UnionType): # optional[T] or similar
        non_none = [a for a in get_args(t) if a is not type(None)]

        if len(non_none) > 1:
            raise TypeError(f"Expected deterministic type; got {non_none}.")
        
        return convert(non_none[0], v)
    
    if v is None: # missing field
        return None
    
    if t is bool:
        return v in ('true', 'True', True)
    
    if o is dict: # mappings
        from .snowflake import Snowflake
        vt = get_args(t)[1]
        return {
            Snowflake(k): convert(vt, x) 
            for k, x in v.items()
        }
    
    if o is list:
        lt = get_args(t)[0]
        return [convert(lt, x) for x in v]
    
    if t is Component:
        from ..api.components import MessageComponentFactory
        return MessageComponentFactory.from_dict(v)

    if hasattr(t, "from_dict"):
        return t.from_dict(v)

    # primitive / fallback
    return t(v)


from dataclasses import dataclass

from ..core.http import HTTPClient

@dataclass
class BaseResource:
    """Represents a Discord Resource object."""

    http: HTTPClient
    """HTTP session for requests."""



from dataclasses import dataclass, field

from ...core.model import DataModel
from ...core.snowflake import Snowflake

@dataclass
class AttachmentModel(DataModel):
    """Represents an attachment object."""

    id: Snowflake
    """Attachment ID."""

    filename: str
    """Name of the file."""

    title: str | None
    """Title of the file."""

    description: str | None
    """Description of the file."""

    content_type: str | None
    """Media type of the file."""

    size: int
    """Size of file (in bytes)."""

    url: str
    """Source URL of the file."""

    proxy_url: str
    """A proxied URL of the file."""

    height: int | None
    """Height of file (if image)."""

    width: int | None
    """Width of file (if image)."""

    ephemeral: bool | None
    """Whether this file is ephemeral."""

    flags: int | None
    """Attachment flags as a combined bitfield."""

@dataclass
class AttachmentPart(DataModel):
    """Represents an attachment."""

    path: str | None = None
    """Relative path to the file."""

    description: str | None = None
    """Description of the file."""

    id: int = field(init=False, default=None)
    """ID of the attachment (internally set)."""

    def to_dict(self):
        return {
            'id': self.id,
            'filename': self.path.split('/')[-1],
            'description': self.description
        }



from dataclasses import dataclass

from ..core.model import DataModel
from ..core.snowflake import Snowflake

from .base_event import Event

from ..api.user import UserModel

@dataclass
class InviteCreateEvent(Event, DataModel):
    """Received when an invite is created."""

    channel_id: Snowflake
    """Channel ID in which the invite belongs."""

    code: str
    """Invite code (unique ID)."""

    guild_id: Snowflake | None
    """Guild ID in which the invite belongs."""

    inviter: UserModel | None
    """User who created invite."""

    uses: int
    """Number of times this invite was used."""

    max_uses: int
    """Max number of times this invite can be used."""

    max_age: int
    """Duration (in seconds) after which this invite expires."""

    temporary: bool
    """Whether this invite only grants temporary membership."""

    created_at: str
    """ISO8601 timestamp for when this invite was created."""


@dataclass
class InviteDeleteEvent(Event, DataModel):
    """Received when an invite is deleted."""

    channel_id: Snowflake
    """Channel ID in which the invite belongs."""

    guild_id: Snowflake | None
    """Guild ID in which the invite belongs."""

    code: str
    """Unique invite code."""



from dataclasses import dataclass, field

from ..core.model import DataModel

@dataclass
class ImageDataPart(DataModel):
    """Represents Discord's data URI scheme for images."""
    
    path: str = None
    """Path to image."""

    def to_dict(self):
        import base64, mimetypes

        mime, _ = mimetypes.guess_type(self.path)
        if mime is None:
            raise ValueError("Unknown file type.")

        with open(self.path, "rb") as f:
            encoded = base64.b64encode(f.read()).decode()

        return f"data:{mime};base64,{encoded}"

@dataclass
class ImageAssetPart(DataModel):
    """Represents fields for creating an image asset."""

    filename: str = None
    """Name of the file."""

    content_type: str = field(init=False)
    """Content type (internally set)."""

    data: bytes = field(init=False)
    """Binary data (internally set)."""

    def to_dict(self):
        import mimetypes

        mime, _ = mimetypes.guess_type(self.filename)
        if mime is None:
            raise ValueError("Unknown file type.")
        
        with open(self.filename, 'rb') as f:
            self.data = f.read()

        return {
            'filename': self.filename,
            'content_type': mime,
            "value": self.data
        }



# scurrypy/etx/components

from .addon import ComponentsAddon
from .ctx import MessageComponentContext, ComponentModalContext

__all__ = [
    "ComponentsAddon",
    
    "MessageComponentContext",
    "ComponentModalContext"
]



from dataclasses import dataclass
from typing import Unpack

from .base_resource import BaseResource

from ..core.snowflake import Snowflake
from ..core.serialization import serialize

from ..bases.channel import GuildChannelCreate

from ..api.guilds.guild import GuildModel, GuildRoleModel
from ..api.guilds.ban import BulkGuildBanPart, GuildBanModel, BulkGuildBanModel
from ..api.guilds.welcome_screen import GuildWelcomeScreenModel
from ..api.guilds.onboarding import GuildOnboadingModel
from ..api.guilds.role import GuildRolePart
from ..api.channels.channel import ChannelModel
from ..api.channels.threads import ActiveThreadsModel
from ..api.messages.sticker import StickerModel, StickerPart

from ..api.invite import InviteModel, InviteWithMetadataModel
from ..api.integration import IntegrationModel
from ..api.user import GuildMemberModel
from ..api.image_data import ImageAssetPart

from ..params.guild import EditGuildRoleParams, EditGuildParams, EditGuildWelcomeScreenParams, EditOnboardingParams, EditGuildStickerParams
from ..params.user import EditGuildMemberParams

@dataclass
class Guild(BaseResource):
    """Represents a Discord guild."""
    
    id: Snowflake
    """ID of the guild."""

    # GUILD
    async def fetch(self, with_counts: bool = False) -> GuildModel:
        """Fetch the Guild object by the given ID.

        Args:
            with_counts (bool, optional): return the approximate member and presence counts for the guild. Defaults to `False`.
            
        Returns:
            (GuildModel): queried guild
        """
        params = {'with_counts': with_counts}

        data = await self.http.request('GET', f'/guilds/{self.id}', params=params)

        return GuildModel.from_dict(data)

    async def edit(self, **options: Unpack[EditGuildParams]) -> GuildModel:
        """Edit this guild.
        Fires [`GuildUpdateEvent`][scurrypy.events.guild_events.GuildUpdateEvent].

        Args:
            options (EditGuildParams): guild with fields to edit

        Returns:
            (GuildModel): edited guild
        """
        options = serialize(options)

        data = await self.http.request('PATCH', f'/guilds/{self.id}', data=options)

        return GuildModel.from_dict(data)

    # --- CHANNELS ---
    async def fetch_channels(self) -> list[ChannelModel]:
        """Fetch this guild's channels.

        !!! note
            Does not include threads!

        Returns:
            (list[ChannelModel]): queried list of the guild's channels
        """
        data = await self.http.request('GET', f'guilds/{self.id}/channels')

        return [ChannelModel.from_dict(channel) for channel in data]
    
    async def fetch_active_threads(self) -> ActiveThreadsModel:
        """Fetch all active threads in a guild (private and public).

        !!! note
            Threads are ordered by their ID in descending order.

        Returns:
            (ActiveThreadsModel): active guild threads
        """
        data = await self.http.request('GET', f'/guilds/{self.id}/threads/active')

        return ActiveThreadsModel.from_dict(data)

    async def create_channel(self, channel: GuildChannelCreate) -> ChannelModel:
        """Create a channel in this guild.
        Fires [`ChannelCreateEvent`][scurrypy.events.channel_events.ChannelCreateEvent].

        !!! important "Permissions"
            Requires `MANAGE_CHANNELS`

        Args:
            channel (GuildChannelCreate): the guild channel to create

        Returns:
            (ChannelModel): created channel
        """
        data = await self.http.request('POST', f'/guilds/{self.id}/channels', data=channel.to_dict())

        return ChannelModel.from_dict(data)

    # --- GUILD MEMBERS ---
    async def fetch_member(self, user_id: Snowflake) -> GuildMemberModel:
        """Fetch a member in this guild.

        !!! warning "Important"
            Requires the `GUILD_MEMBERS` privileged intent!

        Args:
            user_id (Snowflake): user ID of the member to fetch

        Returns:
            (GuildMemberModel): queried guild member
        """
        data = await self.http.request('GET', f'/guilds/{self.id}/members/{user_id}')

        return GuildMemberModel.from_dict(data)

    async def fetch_members(self, limit: int = 1, after: Snowflake = None) -> list[GuildMemberModel]:
        """Fetch guild members in this guild.

        !!! warning "Important"
            Requires the `GUILD_MEMBERS` privileged intent!

        Args:
            limit (int, optional): Max number of members to return Range 1 - 1000. Default `1`.
            after (Snowflake, optional): highest user ID in previous page

        Returns:
            (list[GuildMemberModel]): queried list of guild members
        """
        params = {
            "limit": limit, 
            "after": after
        }

        data = await self.http.request('GET', f'/guilds/{self.id}/members', params=params)

        return [GuildMemberModel.from_dict(member) for member in data]

    async def add_member_role(self, user_id: Snowflake, role_id: Snowflake) -> None:
        """Append a role to a guild member of this guild.
        Fires [`GuildMemberUpdateEvent`][scurrypy.events.user_events.GuildMemberUpdateEvent].

        !!! important "Permissions"
            Requires `MANAGE_ROLES`
        
        Args:
            user_id (Snowflake): ID of the member for the role
            role_id (Snowflake): ID of the role to append
        """
        await self.http.request('PUT', f'/guilds/{self.id}/members/{user_id}/roles/{role_id}')
    
    async def remove_member_role(self, user_id: Snowflake, role_id: Snowflake) -> None:
        """Remove a role from a guild member of this guild.
        Fires [`GuildMemberUpdateEvent`][scurrypy.events.user_events.GuildMemberUpdateEvent].

        !!! important "Permissions"
            Requires `MANAGE_ROLES`

        Args:
            user_id (Snowflake): ID of the member with the role
            role_id (Snowflake): ID of the role to remove
        """
        await self.http.request('DELETE', f'/guilds/{self.id}/members/{user_id}/roles/{role_id}')

    async def search_members(self, query: str = None, limit: int = 1) -> list[GuildMemberModel]:
        """Fetch guild members whose username or nickname starts with the provided query.

        Args:
            query (str, optional): query string to match against
            limit (int, optional): Max number of members to return. Max `1000`. Defaults to `1`.

        Returns:
            list[GuildMemberModel]: queried list of guild members
        """
        data = await self.http.request(
            'GET', 
            f'guild/{self.id}/members/search',
            params={
                'query': query,
                'limit': limit
            }
        )

        return [GuildMemberModel.from_dict(m) for m in data]

    async def edit_member(self, user_id: Snowflake, **options: Unpack[EditGuildMemberParams]) -> GuildMemberModel:
        """Edit a guild member's attributes.
        Fires [`GuildMemberUpdateEvent`][scurrypy.events.user_events.GuildMemberUpdateEvent].

        Args:
            user_id (Snowflake): ID of the member to edit

        Returns:
            (GuildMemberModel): edited guid member
        """
        data = await self.http.request('PATCH', f'/guilds/{self.id}/members/{user_id}', data=options)

        return GuildMemberModel.from_dict(data)

    async def remove_member(self, user_id: Snowflake) -> None:
        """Remove a member from this guild.
        Fires [`GuildMemberRemoveEvent`][scurrypy.events.user_events.GuildMemberRemoveEvent].

        !!! important "Permissions"
            Requires `KICK_MEMBERS`

        Args:
            user_id (Snowflake): ID of the user to kick
        """
        await self.http.request('DELETE', f'/guilds/{self.id}/members/{user_id}')

    # --- BANS ---
    async def fetch_ban(self, user_id: Snowflake) -> GuildBanModel:
        """Fetch a guild ban for the given user ID.

        !!! important "Permissions"
            Requires `BAN_MEMBERS`

        Args:
            user_id (Snowflake): ID of the user to fetch

        Returns:
            (GuildBan): queried ban
        """
        data = self.http.request('GET', f'/guild/{self.id}/bans/{user_id}')

        return GuildBanModel.from_dict(data)

    async def fetch_bans(self, limit: int = 1000, before: Snowflake = None, after: Snowflake = None) -> list[GuildBanModel]:
        """Fetch bans in this guild.

        !!! important "Permissions"
            Requires `BAN_MEMBERS`

        Args:
            limit (int, optional): max number of users to return. Defaults to `1000`.
            before (Snowflake, optional): fetch users before this ID
            after (Snowflake, optional): fetch users after this ID

        Returns:
            (list[GuildBan]): queried list of guild bans
        """
        data = await self.http.request(
            'GET',
            f'/guilds/{self.id}/bans',
            params={
                'limit': limit,
                'before': before,
                'after': after
            }
        )

        return [GuildBanModel.from_dict(i) for i in data]

    async def create_ban(self, user_id: Snowflake, delete_message_seconds: int = 0) -> None:
        """Create a guild ban and optionally delete messages sent by the banned user.
        Fires [`GuildBanAddEvent`][scurrypy.events.guild_events.GuildBanAddEvent].
        
        !!! important "Permissions"
            Requires `BAN_MEMBERS`

        Args:
            user_id (Snowflake): ID of the user to ban
            delete_message_seconds (int, optional): seconds back to delete messages. Max `604800` (7 days). Defaults to `0`.
        """
        await self.http.request(
            'PUT',
            f'/guilds/{self.id}/bans/{user_id}',
            params={'delete_message_seconds': delete_message_seconds}
        )

    async def remove_ban(self, user_id: Snowflake) -> None:
        """Remove the ban for a user.
        Fires [`GuildBanRemoveEvent`][scurrypy.events.guild_events.GuildBanRemoveEvent].

        !!! important "Permissions"
            Requires `BAN_MEMBERS`

        Args:
            user_id (Snowflake): ID of the user in which to remove the ban
        """
        await self.http.request('DELETE', f'/guilds/{self.id}/bans/{user_id}')

    async def bulk_create_ban(self, bulk_ban: BulkGuildBanPart) -> BulkGuildBanModel:
        """Create guild bans and optionally delete messages sent by the banned users.

        !!! important "Permissions"
            Requires `BAN_MEMBERS` and `MANAGE_GUILD`

        Args:
            bulk_ban (BulkGuildBanPart): bulk ban to create
            
        Returns:
            (BulkGuildBanModel): bulk ban response
        """
        data = await self.http.request('POST', f'/guilds/{self.id}/bulk-ban', data=bulk_ban.to_dict())

        return BulkGuildBanModel.from_dict(data)

    # --- ROLES ---
    async def fetch_role_member_counts(self) -> dict:
        """Fetch a map of role IDs to number of members with the role.

        !!! note
            Does not include `@everyone` role.

        Returns:
            (dict): map of role IDs to member count
        """
        return await self.http.request('GET', f'/guilds/{self.id}/roles/member-counts')

    async def fetch_role(self, role_id: Snowflake) -> GuildRoleModel:
        """Fetch a role in this guild.

        Args:
            role_id (Snowflake): ID of the role to fetch

        Returns:
            (GuildRoleModel): queried guild role
        """
        data = await self.http.request('GET', f'/guilds/{self.id}/roles/{role_id}')
        
        return GuildRoleModel.from_dict(data)
    
    async def fetch_roles(self) -> list[GuildRoleModel]:
        """Fetch all roles in this guild.

        Returns:
            (list[GuildRoleModel]): queried list of guild roles
        """
        data = await self.http.request('GET', f'/guilds/{self.id}/roles')
        
        return [GuildRoleModel.from_dict(role) for role in data]

    async def create_role(self, role: GuildRolePart) -> GuildRoleModel:
        """Create a role in this guild.
        Fires [`RoleCreateEvent`][scurrypy.events.role_events.RoleCreateEvent].

        !!! important "Permissions"
            Requires `MANAGE_ROLES`

        Args:
            role (GuildRolePart): fields to create a role

        Returns:
            (GuildRoleModel): created role
        """
        data = await self.http.request('POST', f'/guilds/{self.id}/roles', data=role.to_dict())

        return GuildRoleModel.from_dict(data)

    async def edit_role(self, role_id: Snowflake, **options: Unpack[EditGuildRoleParams]) -> GuildRoleModel:
        """Edit a role in this guild.
        Fires [`RoleUpdateEvent`][scurrypy.events.role_events.RoleUpdateEvent].

        !!! important "Permissions"
            Requires `MANAGE_ROLES`

        Args:
            role_id (Snowflake): ID of role to edit
            options (EditGuildRoleParams): role with fields to edit

        Returns:
            (GuildRoleModel): edited role
        """
        options = serialize(options)

        data = await self.http.request('PATCH', f'/guilds/{self.id}/roles/{role_id}', data=options)

        return GuildRoleModel.from_dict(data)

    async def delete_role(self, role_id: Snowflake) -> None:
        """Delete a role in this guild.
        Fires [`RoleDeleteEvent`][scurrypy.events.role_events.RoleDeleteEvent].

        !!! important "Permissions"
            Requires `MANAGE_ROLES`

        Args:
            role_id (Snowflake): ID of role to delete
        """
        await self.http.request('DELETE', f'/guilds/{self.id}/roles/{role_id}')

    # --- INVITES ---
    async def fetch_invites(self) -> list[InviteModel]:
        """Fetch this guild's invites with no metadata.

        !!! important "Permissions"
            Requires `MANAGE_GUILD` or `VIEW_AUDIT_LOG`

        Returns:
            (list[InviteModel]): queried list of invites without metadata
        """
        data = await self.http.request('GET', f'/guild/{self.id}/invites')

        return [InviteModel.from_dict(i) for i in data]

    async def fetch_invites_with_metadata(self) -> list[InviteWithMetadataModel]:
        """Fetch this guild's invites with metadata.

        !!! important "Permissions"
            Requires `MANAGE_GUILD` and `MANAGE_GUILD` or `VIEW_AUDIT_LOG`

        Returns:
            (list[InviteModel]): queried list of invites with metadata
        """
        data = await self.http.request('GET', f'/guild/{self.id}/invites')

        return [InviteWithMetadataModel.from_dict(i) for i in data]

    # --- INTEGRATIONS ---
    async def fetch_integrations(self) -> list[IntegrationModel]:
        """Fetch this guild's integrations.

        !!! important "Permissions"
            Requires `MANAGE_GUILD`

        Returns:
            (list[IntegrationModel]): queried integrations
        """
        data = await self.http.request('GET', f'/guild/{self.id}/integrations')

        return [IntegrationModel.from_dict(i) for i in data]

    async def delete_integration(self, integration_id: Snowflake) -> None:
        """Delete the attached integration object for this guild.
        Fires [`GuildIntegrationUpdateEvent`][scurrypy.events.integration_events.GuildIntegrationUpdateEvent] 
        and [`GuildIntegrationDeleteEvent`][scurrypy.events.integration_events.GuildIntegrationDeleteEvent].

        !!! important "Permissions"
            Requires `MANAGE_GUILD`

        Args:
            integration_id (Snowflake): ID of the integration to delete
        """
        await self.http.request('DELETE', f'/guilds/{self.id}/integrations/{integration_id}')

    # --- WELCOME SCREEN ---
    async def fetch_welcome_screen(self) -> GuildWelcomeScreenModel:
        """Fetch the welcome screen for this guild.

        !!! important "Permissions"
            Requires `MANAGE_GUILD` if welcome screen is not enabled

        Returns:
            (GuildWelcomeScreenModel): queried welcome screen
        """
        data = await self.http.request('GET', f'/guilds/{self.id}/welcome-screen')

        return GuildWelcomeScreenModel.from_dict(data)

    async def edit_welcome_screen(self, **options: Unpack[EditGuildWelcomeScreenParams]) -> GuildWelcomeScreenModel:
        """Edit this guild's welcome screen.
        May fire [`GuildUpdateEvent`][scurrypy.events.guild_events.GuildUpdateEvent].

        !!! important "Permissions"
            Requires `MANAGE_GUILD`

        Args:
            options (EditGuildWelcomeScreen): fields to edit

        Returns:
            (GuildWelcomeScreenModel): edited welcome screen
        """
        options = serialize(options)

        data = await self.http.request('PATCH', f'/guilds/{self.id}/welcome-screen', data=options)

        return GuildWelcomeScreenModel.from_dict(data)

    # --- ONBOARDING ---
    async def fetch_onboarding(self) -> GuildOnboadingModel:
        """Fetch this guild's onboarding flow.

        Returns:
            (GuildOnboadingModel): queried onboarding flow
        """
        data = await self.http.request('GET', f'/guilds/{self.id}/onboarding')

        return GuildOnboadingModel.from_dict(data)

    async def edit_onboarding(self, **options: Unpack[EditOnboardingParams]) -> GuildOnboadingModel:
        """Modifies this guild's onboarding flow.

        !!! important "Permissions"
            Requires `MANAGE_GUILD` and `MANAGE_ROLES`

        !!! note
            Must be at least **7** Default Channels and at least **5** allow sending message to the @everyone role.
            Constraints depend on the new `mode`.

        Args:
            options (EditOnboardingParams): onboarding field to edit

        Returns:
            (GuildOnboadingModel): edited onboarding flow
        """
        options = serialize(options)
            
        data = await self.http.request(
            'PUT',
            f'/guilds/{self.id}/onboarding',
            params=options
        )
        return GuildOnboadingModel.from_dict(data)

    # --- STICKERS ---
    async def fetch_sticker(self, sticker_id: Snowflake) -> StickerModel:
        """Fetch a sticker from this guild.

        !!! note
            Includes the `user` field if the bot has
            `CREATE_GUILD_EXPRESSIONS` and `MANAGE_GUILD_EXPRESSIONS`

        Args:
            sticker_id (Snowflake): ID of the sticker to fetch

        Returns:
            (StickerModel): queried sticker
        """
        data = await self.http.request('GET', f'/guilds/{self.id}/stickers/{sticker_id}')

        return StickerModel.from_dict(data)

    async def fetch_stickers(self) -> list[StickerModel]:
        """Fetch this guild's stickers.

        !!! note
            Includes the `user` field if the bot has
            `CREATE_GUILD_EXPRESSIONS` and `MANAGE_GUILD_EXPRESSIONS`

        Returns:
            list[StickerModel]: queried guild stickers
        """
        data = await self.http.request('GET', f'/guilds/{self.id}/stickers')

        return [StickerModel.from_dict(i) for i in data]

    async def create_sticker(self, sticker: StickerPart, file: ImageAssetPart) -> StickerModel:
        """Add a sticker to this guild.
        Fires [`GuildStickersUpdateEvent`][scurrypy.events.guild_events.GuildStickersUpdateEvent].

        !!! important "Permissions"
            Requires `CREATE_GUILD_EXPRESSIONS`

        Args:
            sticker (GuildStickerPart): sticker to create
            file (ImageAssetPart): the sticker file to upload
                !!! note
                    Accepted file types: PNG, APNG, GIF, Lottie JSON file.

        Returns:
            (StickerModel): created sticker
        """
        data = await self.http.request(
            'POST', f'/guilds/{self.id}/stickers', 
            data=sticker.to_dict(),
            assets=file.to_dict()
        )
    
        return StickerModel.from_dict(data)

    async def edit_sticker(self, sticker_id: Snowflake, **options: Unpack[EditGuildStickerParams]) -> StickerModel:
        """Edit a sticker from this guild.
        Fires [`GuildStickersUpdateEvent`][scurrypy.events.guild_events.GuildStickersUpdateEvent].

        !!! important "Permissions"
            Requires `CREATE_GUILD_EXPRESSIONS` or `MANAGE_GUILD_EXPRESSIONS`.
            Requires `MANAGE_GUILD_EXPRESSIONS` if not created by the bot.

        Args:
            sticker_id (Snowflake): ID of the sticker to delete
            options (EditGuildStickerParams): fields to edit
        """
        data = await self.http.request('PATCH', f'/guilds/{self.id}/stickers/{sticker_id}', data=options)

        return StickerModel.from_dict(data)

    async def delete_sticker(self, sticker_id: Snowflake) -> None:
        """Delete a sticker from this guild.
        Fires [`GuildStickersUpdateEvent`][scurrypy.events.guild_events.GuildStickersUpdateEvent].

        !!! important "Permissions"
            Requires `CREATE_GUILD_EXPRESSIONS` or `MANAGE_GUILD_EXPRESSIONS`.
            Requires `MANAGE_GUILD_EXPRESSIONS` if not created by the bot.

        Args:
            sticker_id (Snowflake): ID of the sticker to delete
        """
        await self.http.request('DELETE', f'/guilds/{self.id}/stickers/{sticker_id}')



from dataclasses import dataclass, field

from ...core.model import DataModel

from ...bases.components import (
    ContainerChild, 
    SectionAccessoryChild,
    SectionChild,
    Component
)

from ...enums.components import (
    ComponentType,
    SeparatorType
)

@dataclass
class ActionRow(DataModel, Component):
    """Represents a container of interactable components."""

    components: list[Component] | None = None
    """Up to 5 interactive button components or a single select component."""

    type: ComponentType = field(init=False, default=ComponentType.ACTION_ROW)
    """Component type."""

@dataclass
class Section(DataModel, Component, ContainerChild):
    """Represents the Section component.
    
    A Section contextually associates content with an accessory component.
    """

    accessory: Component | None = None
    """A component that is contextually associated to the content of the section.
    
    Supports [`SectionAccessoryChild`][scurrypy.bases.components.SectionAccessoryChild] components.
    """

    components: list[Component] | None = None
    """Component(s) representing the content of the section that is contextually associated to the accessory.
    
    Supports [`SectionChild`][scurrypy.bases.components.SectionChild] components.
    """

    type: ComponentType = field(init=False, default=ComponentType.SECTION)
    """Component type. Always `ComponentType.SECTION` for this class."""


@dataclass
class TextDisplay(DataModel, Component, ContainerChild, SectionChild):
    """Represents the Text Display component.
    
    A Text Display adds markdown formatted text, including mentions (users, roles, etc) and emojis.
    """

    content: str | None = None
    """Text that will be displayed similar to a message."""

    type: ComponentType = field(init=False, default=ComponentType.TEXT_DISPLAY)
    """Component type. Always `ComponentType.TEXT_DISPLAY` for this class."""

@dataclass
class Thumbnail(DataModel, Component, SectionAccessoryChild):
    """Represents the Thumbnail component.
    
    A Thumbnail displays visual media in a small form-factor.
    """
    
    media: str | None = None
    """Media of the thumbnail. http or attachment://<filename> scheme."""

    description: str | None = None
    """Description for the media."""

    spoiler: bool | None = None
    """Whether the thumbnail should be a spoiler (or blurred out). Discord defaults to `False`."""

    type: ComponentType = field(init=False, default=ComponentType.THUMBNAIL)
    """Component type. Always `ComponentType.THUMBNAIL` for this class."""

@dataclass
class MediaGalleryItem(DataModel):
    """Represents the Media Gallery Item component."""

    media: str | None = None
    """Image data. http or attachment://<filename> scheme."""

    description: str | None = None
    """Alt text for the media."""

    spoiler: bool | None = None
    """Whether the thumbnail should be a spoiler (or blurred out). Discord defaults to `False`."""

@dataclass
class MediaGallery(DataModel, Component, ContainerChild):
    """Represents the Media Gallery component.
    
    A Media Gallery displays 1-10 media attachments in an organized gallery format.
    """

    items: list[MediaGalleryItem] | None = None
    """1 to 10 nedia gallery items."""

    type: ComponentType = field(init=False, default=ComponentType.MEDIA_GALLERY)
    """Component type. Always `ComponentType.MEDIA_GALLERY` for this class."""

@dataclass
class File(DataModel, Component, ContainerChild):
    """Represents the File component.
    
    A File displays an uploaded file as an attachment to the message and reference it in the component.
    """

    file: str | None = None
    """File name. ONLY supports attachment://<filename> scheme."""

    spoiler: bool | None = None
    """Whether the thumbnail should be a spoiler (or blurred out). Discord defaults to `False`."""

    type: ComponentType = field(init=False, default=ComponentType.FILE)
    """Component type. Always `ComponentType.FILE` for this class."""

@dataclass
class Separator(DataModel, Component, ContainerChild):
    """Represents the Separator component.
    
    A Separator adds vertical padding and visual division between other components.
    """

    divider: bool | None = None
    """Whether a visual divider should be displayed in the component. Discord defaults to `True`."""

    spacing: SeparatorType | None = None
    """Size of separator padding. Discord defaults to `SeparatorType.SMALL_PADDING`."""

    type: ComponentType = field(init=False, default=ComponentType.SEPARATOR)
    """Component type. Always `ComponentType.SEPARATOR` for this class."""

@dataclass
class Container(DataModel, Component):
    """Represents a container of display and interactable components.
    
    A Container visually encapsulates a collection of components.
    """

    components: list[Component] | None = None
    """Child components that are encapsulated within the Container. 
    
    Supports [`ContainerChild`][scurrypy.bases.components.ContainerChild] components.
    """

    accent_color: int | None = None
    """Color for the accent as an integer."""

    spoiler: bool | None = None
    """If the container should be blurred out. Discord defaults to `False`."""

    type: ComponentType = field(init=False, default=ComponentType.CONTAINER)
    """Component type. Always `ComponentType.CONTAINER` for this class."""

@dataclass
class Label(DataModel, Component):
    """Represents the Discord Label component.
    
    Labels wrap modal components with text as a label and optional description.
    """

    label: str | None = None
    """Label text."""

    component: Component | None = None
    """A component within the label. 
    
    Supports [`LabelChild`][scurrypy.bases.components.LabelChild] components.
    """

    description: str | None = None
    """An optional description text for the label."""

    type: ComponentType = field(init=False, default=ComponentType.LABEL)
    """Component type. Always `ComponentType.LABEL` for this class."""



from dataclasses import dataclass, field

from ...core.model import DataModel

from ...bases.components import (
    LabelChild,
    Component
)

from ...enums.components import (
    ComponentType,
    TextInputStyle
)

@dataclass
class TextInput(DataModel, Component, LabelChild):
    """Represents the Text Input component.
    
    A Text Input allows users to enter free-form text.
    """

    custom_id: str | None = None
    """ID for the input."""

    style: TextInputStyle | None = None
    """Text input style. Discord defaults to `TextInputStyle.SHORT`."""

    min_length: int | None = None
    """Minimum input length for a text input."""

    max_length: int | None = None
    """Maximum input length for a text input."""

    required: bool | None = None
    """Whether this component is required to be filled. Discord defaults to `True`."""

    value: str | None = None
    """Pre-filled value for this component."""

    placeholder: str | None = None
    """Custom placeholder text if the input is empty."""

    type: ComponentType = field(init=False, default=ComponentType.TEXT_INPUT)
    """Component type. Always `ComponentType.TEXT_INPUT` for this class."""


@dataclass
class FileUpload(DataModel, Component, LabelChild):
    """Represents the file upload component.
    
    File Upload allows users to upload files in modals.
    """

    custom_id: str | None = None
    """ID for the file upload."""

    min_values: int | None = None
    """Minimum number of items that must be uploaded. Discord defaults to `1`."""

    max_values: int | None = None
    """Maximum number of items that can be uploaded. Discord defaults to `1`."""

    required: bool | None = None
    """Whether files are required to be uploaded. Discord defaults to `True`."""

    type: ComponentType = field(init=False, default=ComponentType.FILE_UPLOAD)
    """Component type. Always `ComponentType.FILE_UPLOAD` for this class."""

@dataclass
class ListOption(DataModel):
    """Represents an option in a group or checkbox group component."""

    value: str | None = None
    """ID for the option."""

    label: str | None = None
    """User-facing label for the option."""

    description: str | None = None
    """Description for the option."""

    default: bool | None = None
    """Whether to show this option as selected by default."""

@dataclass
class RadioGroup(DataModel, Component, LabelChild):
    """Represents the radio group component.
    
    A Radio Group is for selecting exactly one option from a defined list.
    """

    custom_id: str | None = None
    """ID for the radio group."""

    options: list[ListOption] | None = None
    """List of options to show."""

    required: bool | None = None
    """Whether a selection is required for submission. Discord defaults to `True`."""

    type: ComponentType = field(init=False, default=ComponentType.RADIO_GROUP)
    """Component type. Always `ComponentType.RADIO_GROUP` for this class."""

@dataclass
class CheckboxGroup(DataModel, Component, LabelChild):
    """Represents the checkbox group component.
    
    A Checkbox Group is for selecting one or many options via checkboxes.
    """

    custom_id: str | None = None
    """ID for the checkbox group."""

    options: list[ListOption] | None = None
    """List of options to show."""

    min_values: int | None = None
    """Minimum number of items that must be chosen. Discord defaults to `1`."""

    max_values: int | None = None
    """Maximum number of items that must be chosen. Discord defaults to `1`."""

    required: bool | None = None
    """Whether a selection is required for submission. Discord defaults to the number of options."""

    type: ComponentType = field(init=False, default=ComponentType.CHECKBOX_GROUP)
    """Component type. Always `ComponentType.CHECKBOX_GROUP` for this class."""

@dataclass
class Checkbox(DataModel, Component, LabelChild):
    """Represents a checkbox component.
    
    A Checkbox is for simple yes/no style questions.
    """

    custom_id: str | None = None
    """ID for the option."""

    default: bool | None = None
    """Whether to show this option as selected by default."""

    type: ComponentType = field(init=False, default=ComponentType.CHECKBOX)
    """Component type. Always `ComponentType.CHECKBOX` for this class."""



from dataclasses import dataclass

from ...core.model import DataModel

from ..emoji import EmojiModel

@dataclass
class ReactionCountDetailsModel(DataModel):
    """Represents details for the reaction."""

    burst: int
    """Count of super reactions."""

    normal: int
    """Count of normal reactions."""

@dataclass
class ReactionModel(DataModel):
    """Represents a reaction made."""

    count: int
    """Total number of times this reaction was made."""

    count_details: ReactionCountDetailsModel

    me: bool
    """Whether the bot has reacted with this emoji."""

    me_burst: bool
    """Whether the bot has reacted with a super emoji."""

    emoji: EmojiModel
    """Emoji info."""

    burst_colors: list[str]
    """List of hext colors for the super reaction."""



from dataclasses import dataclass

from ..core.model import DataModel
from ..core.snowflake import Snowflake

from ..enums.integration import IntegrationType

from .application import ApplicationModel

@dataclass
class IntegrationModel(DataModel):
    """Represents a guild integration."""

    id: Snowflake
    """ID of the integration."""

    name: str
    """Name of the integration."""

    type: IntegrationType
    """Type of integration."""

    enabled: bool
    """If the integration is enabled."""

    application: ApplicationModel | None
    """The bot application for Discord integrations."""



# scurrypy/events

from .base_event import Event
from .channel_events import (
    ChannelCreateEvent,
    ChannelUpdateEvent,
    ChannelDeleteEvent,
    ChannelPinsUpdateEvent,
    WebhooksUpdateEvent
)
from .gateway_events import (
    SessionStartLimit,
    GatewayEvent
)
from .guild_events import (
    GuildCreateEvent,
    GuildUpdateEvent,
    GuildDeleteEvent,

    GuildBanAddEvent,
    GuildBanRemoveEvent,

    GuildEmojisUpdateEvent,
    GuildStickersUpdateEvent
)
from .integration_events import (
    GuildIntegrationCreateEvent,
    GuildIntegrationUpdateEvent,
    GuildIntegrationsUpdateEvent,
    GuildIntegrationDeleteEvent
)
from .interaction_events import InteractionEvent
from .invite_events import (
    InviteCreateEvent, 
    InviteDeleteEvent
)
from .message_events import (
    MessageCreateEvent,
    MessageUpdateEvent,
    MessageDeleteEvent,
    BulkMessageDeleteEvent
)
from .reaction_events import (
    ReactionAddEvent,
    ReactionRemoveEvent,
    ReactionRemoveEmojiEvent,
    ReactionRemoveAllEvent
)
from .ready_event import ReadyEvent
from .role_events import (
    RoleCreateEvent,
    RoleUpdateEvent,
    RoleDeleteEvent
)
from .thread_events import (
    ThreadCreateEvent,
    ThreadUpdateEvent,
    ThreadDeleteEvent,
    ThreadMemberUpdateEvent,
    ThreadMembersUpdateEvent,
    ThreadListSyncEvent
)
from .user_events import (
    UserUpdateEvent,
    GuildMemberAddEvent,
    GuildMemberUpdateEvent,
    GuildMemberRemoveEvent
)

__all__ = [
    "Event",

    "ChannelCreateEvent",
    "ChannelUpdateEvent",
    "ChannelDeleteEvent",
    "ChannelPinsUpdateEvent",
    "WebhooksUpdateEvent",

    "SessionStartLimit",
    "GatewayEvent",

    "GuildCreateEvent",
    "GuildUpdateEvent",
    "GuildDeleteEvent",

    "GuildBanAddEvent",
    "GuildBanRemoveEvent",
    "GuildEmojisUpdateEvent",
    "GuildStickersUpdateEvent",

    "GuildIntegrationCreateEvent",
    "GuildIntegrationUpdateEvent",
    "GuildIntegrationsUpdateEvent",
    "GuildIntegrationDeleteEvent",

    "InteractionEvent",

    "InviteCreateEvent", 
    "InviteDeleteEvent",

    "MessageCreateEvent",
    "MessageUpdateEvent",
    "MessageDeleteEvent",
    "BulkMessageDeleteEvent",

    "ReactionAddEvent",
    "ReactionRemoveEvent",
    "ReactionRemoveEmojiEvent",
    "ReactionRemoveAllEvent",

    "ReadyEvent",

    "RoleCreateEvent",
    "RoleUpdateEvent",
    "RoleDeleteEvent",

    "ThreadCreateEvent",
    "ThreadUpdateEvent",
    "ThreadDeleteEvent",
    "ThreadMemberUpdateEvent",
    "ThreadMembersUpdateEvent",
    "ThreadListSyncEvent",

    "UserUpdateEvent",
    "GuildMemberAddEvent",
    "GuildMemberUpdateEvent",
    "GuildMemberRemoveEvent"
]



from .enum_types import DiscordFlags

class GuildMemberFlags(DiscordFlags):
    DID_REJOIN = 1 << 0
    COMPLETED_ONBOARDING = 1 << 1
    BYPASSES_VERIFICATION = 1 << 2
    STARTED_ONBOARDING = 1 << 3
    STARTED_HOME_ACTIONS = 1 << 5
    COMPLETED_HOME_ACTIONS = 1 << 6
    AUTOMOD_QUARANTINED_USERNAME = 1 << 7
    AUTOMOD_QUARANTINED_TAG = 1 << 8



from enum import IntFlag, IntEnum, StrEnum

class DiscordTypes(IntEnum):
    """Base class for Discord's types."""
    @classmethod
    def from_dict(cls, v):
        return cls(int(v))
    
class DiscordFlags(IntFlag):
    """Base class for Discord's flags."""
    @classmethod
    def from_dict(cls, v):
        return cls(int(v))

class DiscordString(StrEnum):
    """Base class for Discord's pre-defined strings."""
    @classmethod
    def from_dict(cls, v):
        return cls(str(v))



from ..events import *
from ..enums.events import EventType

EVENTS = {
    # startup events
    EventType.READY: ReadyEvent,

    # channel events
    EventType.CHANNEL_CREATE: ChannelCreateEvent,
    EventType.CHANNEL_UPDATE: ChannelUpdateEvent,
    EventType.CHANNEL_DELETE: ChannelDeleteEvent,
    
    EventType.CHANNEL_PINS_UPDATE: ChannelPinsUpdateEvent,

    EventType.THREAD_CREATE: ThreadCreateEvent,
    EventType.THREAD_UPDATE: ThreadUpdateEvent,
    EventType.THREAD_DELETE: ThreadDeleteEvent,
    EventType.THREAD_MEMBER_UPDATE: ThreadMemberUpdateEvent,
    EventType.THREAD_MEMBERS_UPDATE: ThreadMembersUpdateEvent,
    EventType.THREAD_LIST_SYNC: ThreadListSyncEvent,

    EventType.BULK_MESSAGE_DELETE: BulkMessageDeleteEvent,
    
    EventType.WEBHOOKS_UPDATE: WebhooksUpdateEvent,

    # invite events
    EventType.INVITE_CREATE: InviteCreateEvent,
    EventType.INVITE_DELETE: InviteDeleteEvent,

    # guild events
    EventType.GUILD_CREATE: GuildCreateEvent,
    EventType.GUILD_UPDATE: GuildUpdateEvent,
    EventType.GUILD_DELETE: GuildDeleteEvent,

    EventType.GUILD_MEMBER_ADD: GuildMemberAddEvent,
    EventType.GUILD_MEMBER_UPDATE: GuildMemberUpdateEvent,
    EventType.GUILD_MEMBER_REMOVE: GuildMemberRemoveEvent,

    EventType.GUILD_EMOJIS_UPDATE: GuildEmojisUpdateEvent,

    EventType.GUILD_STICKERS_UPDATE: GuildStickersUpdateEvent,

    EventType.GUILD_BAN_ADD: GuildBanAddEvent,
    EventType.GUILD_BAN_REMOVE: GuildBanRemoveEvent,

    # integration events
    EventType.INTEGRATION_CREATE: GuildIntegrationCreateEvent,
    EventType.GUILD_INTEGRATIONS_UPDATE: GuildIntegrationsUpdateEvent,
    EventType.INTEGRATION_UPDATE: GuildIntegrationUpdateEvent,
    EventType.INTEGRATION_DELETE: GuildIntegrationDeleteEvent,

    # interaction events
    EventType.INTERACTION_CREATE: InteractionEvent,

    # message events
    EventType.MESSAGE_CREATE: MessageCreateEvent,
    EventType.MESSAGE_UPDATE: MessageUpdateEvent,
    EventType.MESSAGE_DELETE: MessageDeleteEvent,

    # reaction events
    EventType.MESSAGE_REACTION_ADD: ReactionAddEvent,
    EventType.MESSAGE_REACTION_REMOVE: ReactionRemoveEvent,
    EventType.MESSAGE_REACTION_REMOVE_ALL: ReactionRemoveAllEvent,
    EventType.MESSAGE_REACTION_REMOVE_EMOJI: ReactionRemoveEmojiEvent,

    # role events
    EventType.ROLE_CREATE: RoleCreateEvent,
    EventType.ROLE_UPDATE: RoleUpdateEvent,
    EventType.ROLE_DELETE: RoleDeleteEvent
}



import sys
import importlib.metadata

try:
    __version__ = importlib.metadata.version("scurrypy")
except importlib.metadata.PackageNotFoundError:
    # Fallback for development/editable installs
    __version__ = "0.0.0-dev"

__title__ = "scurrypy"
__url__ = "https://github.com/scurry-works/scurrypy"

USER_AGENT = (
    f"DiscordBot ({__url__}, {__version__}) "
    f"{__title__}/{__version__} "
    f"Python/{sys.version_info.major}.{sys.version_info.minor}"
)

GATEWAY_PROPERTIES = {
    "$os": sys.platform,
    "$browser": __title__,
    "$device": __title__
}

# silence internals by default
import logging

logging.getLogger('websockets').setLevel(logging.WARNING)
logging.getLogger('aiohttp').setLevel(logging.WARNING)
logging.getLogger('asyncio').setLevel(logging.WARNING)



from dataclasses import dataclass

from ...core.model import DataModel
from ...core.snowflake import Snowflake

from ...enums.permissions import Permissions
from ..image_data import ImageDataPart

@dataclass
class GuildRoleColorModel(DataModel):
    """Represents role color data."""

    primary_color: int
    """Primary color of the role."""

    secondary_color: int
    """Secondary color of the role. Creates a gradient."""

    tertiary_color: int
    """Tertiary color of the role. Creates a holographic style."""

@dataclass
class GuildRoleModel(DataModel):
    """Represents a Discord role."""

    id: Snowflake
    """ID of the role."""

    name: str
    """Name of the role."""

    colors: GuildRoleColorModel
    """Colors of the role."""

    hoist: bool
    """If the role is pinned in user listing."""

    position: int
    """Position of the role."""

    permissions: int
    """Permission bit set. [INT_LIMIT]"""

    managed: bool
    """If the role is managed by an integration."""

    mentionable: bool
    """If the role is mentionable."""

    flags: int
    """Role flags combined as a bitfield."""

    icon: str | None
    """Icon hash of the role."""

    permissions: str
    """permission bits set."""

    unicode_emoji: str | None
    """Unicode emoji of the role."""

@dataclass
class GuildRoleColorsPart(DataModel):
    """Parameters for setting role colors."""

    primary_color: int | None = None
    """Primary color of the role."""

    secondary_color: int | None = None
    """Secondary color of the role. Creates a gradient."""

    tertiary_color: int | None = None
    """Tertiary color of the role. Creates a holographic style."""

@dataclass
class GuildRolePart(DataModel):
    """Parameters for creating a role."""

    name: str | None = None
    """Name of the role. Discord defaults to \"user role\"."""

    colors: GuildRoleColorsPart | None = None
    """Colors of the role. Discord defaults to primary color set to `0`."""

    icon: ImageDataPart | None = None
    """Icon of the role (if guild has `ROLE_ICONS` feature)."""

    permissions: Permissions | None = None
    """Permission bit set. [`INT_LIMIT`]"""

    hoist: bool | None = None
    """If the role is pinned in the user listing. Discord defaults to `False`."""

    mentionable: bool | None = None
    """If the role is mentionable. Discord defaults to `False`."""

    unicode_emoji: str | None = None
    """Unicode emoji of the role."""



from .enum_types import DiscordFlags, DiscordTypes

class ChannelType(DiscordTypes):
    """Represents the types of channels."""

    GUILD_TEXT = 0
    """Text channel within a guild."""

    GUILD_CATEGORY = 4
    """Organizational category for channels."""

    GUILD_ANNOUNCEMENT = 5
    """Channel users can follow and crosspost into their own server."""

    ANNOUNCEMENT_THREAD = 10
    """Temporary sub-channel within a `GUILD_ANNOUNCEMENT` channel."""

    PUBLIC_THREAD = 11
    """Temporary sub-channel within a `GUILD_TEXT` or `GUILD_FORUM` channel."""

    PRIVATE_THREAD = 12
    """Temporary sub-channel within a `GUILD_TEXT` channel only viewable by invitees and members with `MANAGE_THREADS`."""

    GUILD_DIRECTORY = 14
    """Channel in a hub containing the listed servers."""

    GUILD_FORUM = 15
    """Channel that can only contain threads."""

class ChannelFlags(DiscordFlags):
    """Represents constant bit fields for channel flags."""

    PINNED = 1 << 1
    """This thread is pinned to the top of its parent `GUILD_FORUM` channel."""

    REQUIRE_TAG = 1 << 4
    """Whether a tag is required when creating a thread in a `GUILD_FORUM` channel."""

class SortOrderType(DiscordTypes):
    """Represents sort order types for `GUILD_FORUM` channels."""

    LATEST_ACTIVITY = 0
    """Sort by activity."""

    CREATION_DATE = 1
    """Sort by creation time (recent to oldest)."""

class ForumLayoutType(DiscordTypes):
    """Represents `GUILD_FORUM` layout types."""

    NOT_SET = 0
    """No default layout has been set."""

    LIST_VIEW = 1
    """Display posts as a list."""

    GALLERY_VIEW = 2
    """Display posts as a collection of tiles."""

class AutoArchiveDurationType(DiscordTypes):
    """Auto archive duration options."""
    ONE_HR = 60
    ONE_DAY = 1440
    THREE_DAYS = 4320
    ONE_WEEK = 10080



# scurrypy/ext/cache

from .application_emojis import ApplicationEmojisCacheAddon
from .channels import GuildChannelCacheAddon
from .guild_emojis import GuildEmojiCacheAddon
from .guild_roles import GuildRoleCacheAddon

__all__ = [
    "ApplicationEmojisCacheAddon",
    "GuildChannelCacheAddon",
    "GuildEmojiCacheAddon",
    "GuildRoleCacheAddon"
]



from dataclasses import dataclass, fields

from .serialization import convert, serialize

@dataclass
class DataModel:    
    """DataModel is a base class for Discord JSONs that provides 
        hydration from raw dicts, and optional field defaults.
    """

    @classmethod
    def from_dict(cls, data: dict):
        """Hydrates the given data into the dataclass.

        Args:
            data (dict): the JSON data

        Returns:
            (cls): hydrated dataclass
        """
        if data is None:
            return None
        
        kwargs = {}

        for f in fields(cls):
            if not f.init: # ignore init=False fields
                continue

            kwargs[f.name] = convert(f.type, data.get(f.name))

        return cls(**kwargs)
        
    def to_dict(self):
        """Recursively turns the dataclass into a dictionary and drops empty fields.

        Returns:
            (dict): serialized dataclasss
        """
        result = {}
        for f in fields(self):
            if f.name.startswith('_'):
                continue
            val = getattr(self, f.name)
            # only include real values
            if val is not None:
                result[f.name] = serialize(val)
        return result



from scurrypy import Client
from scurrypy.bases import Addon
from scurrypy.enums import EventType
from scurrypy.core import Snowflake, DiscordError
from scurrypy.api.channels import ChannelModel
from scurrypy.events import GuildCreateEvent, GuildDeleteEvent, ChannelCreateEvent, ChannelUpdateEvent, ChannelDeleteEvent

class GuildChannelCacheAddon(Addon):
    """Defines caching channels and lookup."""

    def __init__(self, client: Client):
        self.bot = client

        self.channels: dict[Snowflake, dict[Snowflake, ChannelModel]] = {}  # stores OBJECTS
        self.channel_index: dict[Snowflake, ChannelModel] = {}        # stores REFERENCES

        client.add_event_listener(EventType.GUILD_CREATE, self.on_guild_create)
        client.add_event_listener(EventType.GUILD_DELETE, self.on_guild_delete)

        client.add_event_listener(EventType.CHANNEL_CREATE, self.on_channel_create)
        client.add_event_listener(EventType.CHANNEL_UPDATE, self.on_channel_update)
        client.add_event_listener(EventType.CHANNEL_DELETE, self.on_channel_delete)

    def on_guild_create(self, event: GuildCreateEvent):
        """Append new guild channels to cache. Also add channels to index.

        Args:
            event (GuildCreateEvent): the GUILD_CREATE event
        """
        guild_dict = self.channels.setdefault(event.id, {})

        for ch in event.channels:
            guild_dict[ch.id] = ch
            self.channel_index[ch.id] = ch

    def on_guild_delete(self, event: GuildDeleteEvent):
        """Remove guild channels from cache. Also remove channels from index

        Args:
            event (GuildDeleteEvent): the GUILD_DELETE event
        """
        removed_channels = self.channels.pop(event.id, {})

        for ch in removed_channels.values():
            self.channel_index.pop(ch.id, None)

    def on_channel_create(self, event: ChannelCreateEvent):
        """Append channel to guild key. Also append channel to index.

        Args:
            event (GuildChannelCreateEvent): the CHANNEL_CREATE event
        """
        model = ChannelModel.from_dict(event.raw)
        guild_dict = self.channels.setdefault(event.guild_id, {})

        guild_dict[event.id] = model
        self.channel_index[event.id] = model

    def on_channel_update(self, event: ChannelUpdateEvent):
        """Replace channel in guild key. Also replace channel in index.

        Args:
            event (GuildChannelUpdateEvent): the CHANNEL_UPDATE event
        """
        model = ChannelModel.from_dict(event.raw)
        guild_dict = self.channels.setdefault(event.guild_id, {})

        guild_dict[event.id] = model
        self.channel_index[event.id] = model

    def on_channel_delete(self, event: ChannelDeleteEvent):
        """Remove channel from guild key. Also remove channel from index.

        Args:
            event (GuildChannelDeleteEvent): the CHANNEL_DELETE event
        """
        model = self.channel_index.pop(event.id, None)
        if model:
            self.channels.get(event.guild_id, {}).pop(event.id, None)

    async def get_channel(self, channel_id: Snowflake):
        """Fetch a guild channel. If not found, request and store it.

        Args:
            channel_id (Snowflake): ID of channel

        Returns:
            (ChannelModel | None): hydrated channel object or None if fetch failed
        """
        channel = self.channel_index.get(channel_id)
        if channel:
            return channel

        try:
            channel = await self.bot.channel(channel_id).fetch()
        except DiscordError:
            return None

        self.put(channel)

        return channel

    def put(self, channel: ChannelModel):
        """Put a new channel into the cache.

        Args:
            channel (ChannelModel): the channel object

        Raises:
            ValueError: missing `guild_id`
        """
        if channel.guild_id is None:
            raise ValueError("Cannot cache a channel without a guild_id.")
        
        guild_dict = self.channels.setdefault(channel.guild_id, {})
        guild_dict[channel.id] = channel
        self.channel_index[channel.id] = channel



from .enums.enum_types import DiscordFlags

class Intents(DiscordFlags):
    """Gateway intent flags (bitwise).  
    
    For an exhaustive list what intents let your bot listen to what events, 
    see the [list of intents](https://discord.com/developers/docs/events/gateway#list-of-intents).

    !!! note
        Not all intents are listed. Intents not listed are not yet supported.
    """

    GUILDS = 1 << 0
    """Receive events related to guilds."""

    GUILD_MEMBERS = 1 << 1
    """Receive events related to guild members.

    !!! warning "Privileged Intent"
        Requires the app setting `Server Members Intent` to be toggled.
    """

    GUILD_MODERATION = 1 << 2
    """Receive events related to guild moderation."""

    GUILD_EXPRESSIONS = 1 << 3
    """Receive events related to custom emojis and stickers."""

    GUILD_INTEGRATIONS = 1 << 4
    """Receive events related to integrations within a guild."""

    GUILD_WEBHOOKS = 1 << 5
    """Receive events related to webhooks."""

    GUILD_INVITES = 1 << 6
    """Receive events related to creating guild invites."""

    GUILD_MESSAGES = 1 << 9
    """Receive events about messages within a guild."""

    GUILD_MESSAGE_REACTIONS = 1 << 10
    """Track changes in reactions on messages."""

    DIRECT_MESSAGES = 1 << 12
    """Receive events related to DMs."""

    DIRECT_MESSAGE_REACTIONS = 1 << 13
    """Receive events related to DM reactions."""

    MESSAGE_CONTENT = 1 << 15
    """Access content of messages.

    !!! warning "Privileged Intent"
        Requires the app setting `Message Content Intent` to be toggled.
    """

    DEFAULT = GUILDS | GUILD_MESSAGES



from dataclasses import dataclass

from ...core.model import DataModel
from ...core.snowflake import Snowflake

from ..user import UserModel

@dataclass
class GuildBanModel(DataModel):
    """Represents the guild ban object."""

    reason: str
    """Reason for the ban."""
    
    user: UserModel
    """Banned user object."""

@dataclass
class BulkGuildBanModel(DataModel):
    """Response body for creating bulk guild bans."""

    banned_users: list[Snowflake]
    """IDs of successfully banned users."""

    failed_users: list[Snowflake]
    """IDs of users not banned."""

@dataclass
class BulkGuildBanPart(DataModel):
    """Represents fields for creating a bulk ban."""

    user_ids: list[Snowflake] = None
    """List of user IDs to ban. Max `200`."""

    delete_message_seconds: int | None = None
    """seconds back to delete messages. Max `604800` (7 days). Discord defaults to `0`."""



from dataclasses import dataclass
from typing import Unpack

from .base_resource import BaseResource

from ..core.snowflake import Snowflake
from ..core.serialization import serialize

from ..api.channels.channel import ChannelModel

from ..api.user import UserModel, GuildMemberModel

from ..params.user import EditUserParams

@dataclass
class User(BaseResource):
    """A Discord user."""

    async def fetch(self, user_id: Snowflake) -> UserModel:
        """Fetch this user by ID.

        Args:
            user_id (Snowflake): ID of user to fetch

        Returns:
            (UserModel): queried user
        """
        data = await self.http.request('GET', f'/users/{user_id}')

        return UserModel.from_dict(data)

    async def fetch_guild_member(self, guild_id: Snowflake, user_id: Snowflake) -> GuildMemberModel:
        """Fetch this user's guild member data.

        Args:
            guild_id (Snowflake): ID of guild to fetch data from
            user_id (Snowflake): ID of user to fetch

        Returns:
            (GuildMemberModel): queried guild member for the user
        """
        data = await self.http.request('GET', f'/guilds/{guild_id}/members/{user_id}')

        return GuildMemberModel.from_dict(data)

    async def modify_current_user(self, **options: Unpack[EditUserParams]) -> UserModel:
        """Modify the bot's account settings.
        Fires [`UserUpdateEvent`][scurrypy.events.user_events.UserUpdateEvent].

        Args:
            options (EditUserParams): fields to edit

        Returns:
            (UserModel): edited user
        """
        options = serialize(options)

        data = await self.http.request('PATCH', '/users/@me', data=options)

        return UserModel.from_dict(data)

    async def leave_guild(self, guild_id: Snowflake) -> None:
        """Make the bot leave a guild.
        Fires [`GuildDeleteEvent`][scurrypy.events.guild_events.GuildDeleteEvent]
        and [`GuildMemberRemoveEvent`][scurrypy.events.user_events.GuildMemberRemoveEvent].

        Args:
            guild_id (Snowflake): ID of the guild to leave
        """
        await self.http.request('DELETE', f'/users/@me/guilds/{guild_id}')

    async def create_dm(self, user_id: Snowflake) -> ChannelModel:
        """Create a DM between the bot and this user.

        Args:
            user_id (Snowflake): ID of user to create DM with
        
        Returns:
            (ChannelModel): created or existing DM channel
        """
        data = await self.http.request(
            'POST', 
            '/users/@me/channels', 
            data={'recipient_id': user_id}
        )

        return ChannelModel.from_dict(data)



from .enum_types import DiscordTypes, DiscordString

class ComponentType(DiscordTypes):
    ACTION_ROW = 1
    BUTTON = 2
    STRING_SELECT = 3
    TEXT_INPUT = 4
    USER_SELECT = 5
    ROLE_SELECT = 6
    MENTIONABLE_SELECT = 7
    CHANNEL_SELECT = 8
    SECTION = 9
    TEXT_DISPLAY = 10
    THUMBNAIL = 11
    MEDIA_GALLERY = 12
    FILE = 13
    SEPARATOR = 14
    CONTAINER = 17
    LABEL = 18
    FILE_UPLOAD = 19
    RADIO_GROUP = 21
    CHECKBOX_GROUP = 22
    CHECKBOX = 23

class SeparatorType(DiscordTypes):
    """Represents separator types constants."""

    SMALL_PADDING = 1
    """Small separator padding."""
    
    LARGE_PADDING = 2
    """Large separator padding."""

class ButtonStyle(DiscordTypes):
    """Represents button styles for a Button component."""

    PRIMARY = 1
    """The most important or recommended action in a group of options. (Blurple)"""

    SECONDARY = 2
    """Alternative or supporting actions. (Gray)"""

    SUCCESS = 3
    """Positive confirmation or completion actions. (Green)"""

    DANGER = 4
    """An action with irreversible consequences. (Red)"""

    LINK = 5
    """Navigates to a URL. (Gray + window)"""

class TextInputStyle(DiscordTypes):
    """Represents the types of Text Inputs."""

    SHORT = 1
    """One line text input."""

    PARAGRAPH = 2
    """Multi-line text input."""

class DefaultValueType(DiscordString):
    ROLE = "role"
    CHANNEL = "channel"
    USER = "user"



class InteractionData:
    """Marker class for all interaction data events.
    
    !!! tip "Children"
        [`ApplicationCommandDataModel`][scurrypy.api.interactions.ApplicationCommandDataModel]
        [`MessageComponentDataModel`][scurrypy.api.interactions.MessageComponentDataModel]
        [`ModalDataModel`][scurrypy.api.interactions.ModalDataModel]
    """
    __slots__ = ()



from dataclasses import dataclass

from ...core.model import DataModel
from ...core.snowflake import Snowflake

@dataclass
class GuildWelcomeChannelModel(DataModel):
    """Represents channels shown on a welcome screen."""

    channel_id: Snowflake
    """ID of the channel."""

    description: str
    """Description for the channel."""

    emoji_id: Snowflake
    """Emoji ID for the welcome screen (if custom)."""

    emoji_name: str
    """Emoji name for the welcome screen."""

@dataclass
class GuildWelcomeScreenModel(DataModel):
    """Represents a guild's welcome screen."""

    description: str
    """Guild description displayed."""

    welcome_channels: list[GuildWelcomeChannelModel]
    """Channels displayed on the welcome screen. Max `5`."""

@dataclass
class WelcomeScreenChannelPart(DataModel):
    """Represents fields for creating a welcome screen channel."""

    channel_id: Snowflake | None = None
    """ID of the channel to display."""

    description: str | None = None
    """Description for the channel to display."""

    emoji_id: Snowflake | None = None
    """ID of the emoji (if custom)."""

    emoji_name: str | None = None
    """Name of the emoji."""



from dataclasses import dataclass, field

from ...core.model import DataModel

from ...enums.command import CommandOptionType, CommandType

@dataclass
class CommandOptionChoicePart(DataModel):
    """Choice for a command option."""

    name: str | None = None
    """Name of the choice."""

    value: str | int | float | None = None
    """Value for the user to select (same as option type)."""

    name_localizations: dict | None = None
    """Dictionary with keys in available locales."""

@dataclass
class CommandOptionPart(DataModel):
    """Option for a slash command."""

    type: CommandOptionType | None = None
    """Type of option."""

    name: str | None = None
    """Name of option."""

    description: str | None = None
    """Description of option."""

    required: bool | None = None
    """Whether this option is required. Discord defaults to `False`."""

    choices: list[CommandOptionChoicePart] | None = None
    """Choices for the user to pick from, max 25. Only valid for STRING, INTEGER, NUMBER option types."""

    autocomplete: bool | None = None
    """Whether autocomplete interactions are enabled for this option. Discord defaults to `False`."""

@dataclass
class SlashCommandPart(DataModel):
    """Represents the slash command object."""

    name: str | None = None
    """Name of the command."""

    description: str | None = None
    """Description of the command."""

    options: list[CommandOptionPart] | None = None
    """Parameters or options for the command."""

    type: CommandType = field(init=False, default=CommandType.CHAT_INPUT)
    """Command type. Always `CommandType.CHAT_INPUT` for this class."""



# scurrypy/api/messages

from .attachment import (
    AttachmentModel, 
    AttachmentPart
)
from .embed import (
    EmbedAuthor, 
    EmbedThumbnail, 
    EmbedField, 
    EmbedImage, 
    EmbedFooter, 
    Embed
)
from .message import (
    MessageModel, 
    PinnedMessageModel, 
    MessageReferencePart, 
    MessagePart
)
from .reaction import (
    ReactionCountDetailsModel, 
    ReactionModel
)
from .sticker import (
    StickerModel, 
    StickerItemModel, 
    StickerPackModel,
    StickerPart
)

__all__ = [
    "AttachmentModel", 
    "AttachmentPart",

    "EmbedAuthor", 
    "EmbedThumbnail", 
    "EmbedField", 
    "EmbedImage", 
    "EmbedFooter", 
    "Embed",

    "MessageModel", 
    "PinnedMessageModel", 
    "MessageReferencePart", 
    "MessagePart",

    "ReactionCountDetailsModel", 
    "ReactionModel",

    "StickerModel", 
    "StickerItemModel", 
    "StickerPackModel",
    "StickerPart"
]



from dataclasses import dataclass

from ...core.model import DataModel
from ...core.snowflake import Snowflake

from ...enums.permissions import Permissions
from ...enums.command import CommandOptionType, CommandType

@dataclass
class ApplicationCommandOptionChoiceModel(DataModel):
    """Represents the application command option choice object."""

    name: str
    """Name of the choice."""

    value: str
    """Value for the choice.
    
    !!! note
        Convert based on expected type (str, int or double)
    """

@dataclass
class ApplicationCommandOptionModel(DataModel):
    """Represents the application command option object."""

    type: CommandOptionType
    """Type of command option."""

    name: str
    """Name of the command option."""

    descripton: str
    """Description for the command option."""
    
    required: bool | None
    """Whether this option is required. Discord defaults to `False`."""

    choices: list[ApplicationCommandOptionChoiceModel] | None
    """Choices for the user to pick from."""

    channel_types: list[int] | None
    """Channels shown will be restricted to these types."""

    min_value: int | None
    """Minimum value allowed."""

    max_value: int | None
    """Maximum value allowed."""

    min_length: int | None
    """Minimum length allowed."""

    max_length: int | None
    """Maximum length allowed."""

    autocomplete: bool | None
    """Whether autocomplete interactions are enabled for this option."""

@dataclass
class ApplicationCommandModel(DataModel):
    """Represents the application command object."""

    id: Snowflake
    """Unique ID of command."""

    type: CommandType | None
    """Type of command. Discord defaults to `ApplicationCommandTypes.CHAT_INPUT`."""

    application_id: Snowflake
    """ID of the parent application."""

    guild_id: Snowflake | None
    """Guild ID of the command, if not global."""

    name: str
    """Name of the command."""

    description: str
    """Description for `CHAT_INPUT` commands. 
    
    !!! note
        Empty for `USER` and `MESSAGE` commands.
    """

    options: list[ApplicationCommandOptionModel] | None
    """Parameters for the command."""

    default_member_permissions: Permissions
    """Set of permissions represented as a bit set. [`INT_LIMIT`]"""

    nsfw: bool | None
    """Whether the command is age-restricted. Discord defaults to `False`."""



from datetime import datetime, timezone

DISCORD_EPOCH = 1420070400000

class Snowflake(int):
    """Represents a Discord snowflake ID."""

    @property
    def timestamp_ms(self) -> int:
        """Timestamp associated with the ID as an integer in milliseconds (ms)."""
        return (self >> 22) + DISCORD_EPOCH

    @property
    def datetime(self) -> datetime:
        """Timestamp associated with the ID as a `datetime` object."""
        return datetime.fromtimestamp(self.timestamp_ms / 1000, tz=timezone.utc)

    @property
    def isoformat(self) -> str:
        """Timestamp associated with the ID as a string in ISO8601 format."""
        return self.datetime.isoformat()

    @property
    def worker_id(self) -> int:
        """Worker ID associated with the ID."""
        return (self & 0x3E0000) >> 17

    @property
    def process_id(self) -> int:
        """Process ID associated with the ID."""
        return (self & 0x1F000) >> 12

    @property
    def increment(self) -> int:
        """Increment associated with the ID."""
        return self & 0xFFF

    @classmethod
    def from_dict(cls, v):
        return cls(int(v))



from dataclasses import dataclass

from ..core.model import DataModel
from ..core.snowflake import Snowflake

from .base_event import Event

from ..api.user import UserModel, GuildMemberModel

@dataclass
class UserUpdateEvent(Event, UserModel):
    """Received when a user's settings are updated."""
    pass

@dataclass
class GuildMemberAddEvent(Event, GuildMemberModel):
    """Received when a member joins a guild the bot is in.

    !!! warning
        Requires privileged `GUILD_MEMBERS` intent.
    """

    guild_id: Snowflake
    """ID of the guild."""

@dataclass
class GuildMemberUpdateEvent(Event, DataModel):
    """Received when a guild member is updated.
    
    !!! warning
        Requires privileged `GUILD_MEMBERS` intent.
    """
    guild_id: Snowflake
    """ID of the guild."""

    roles: list[Snowflake]
    """List of user's roles (their IDs)."""

    user: UserModel
    """The User object."""

    avatar: str
    """Guild avatar hash."""

    banner: str
    """Guild banner hash."""

    joined_at: str
    """When the user joined the guild"""

@dataclass
class GuildMemberRemoveEvent(Event, DataModel):
    """Received when a member leaves or is kicked/banned from a guild the bot is in.
    
    !!! warning
        Requires privileged `GUILD_MEMBERS` intent.
    """

    guild_id: Snowflake
    """ID of the guild."""

    user: UserModel
    """User object of the user leaving the guild."""
